from typing import Dict, List

from .name2idx import C, V


class ReactionNetwork(object):
    def __init__(self) -> None:
        super(ReactionNetwork, self).__init__()
        self.reactions: Dict[str, List[int]] = {
            "all": [i for i in range(1, 11)],
        }

    @staticmethod
    def flux(t, y, x) -> dict:
        v = {}
        v[1] = (
            x[C.V1]
            * y[V.MKKK]
            / ((1 + (y[V.MAPK_PP] / x[C.KI]) ** x[C.n]) * (x[C.K1] + y[V.MKKK]))
        )
        v[2] = x[C.V2] * y[V.MKKK_P] / (x[C.K2] + y[V.MKKK_P])
        v[3] = x[C.k3] * y[V.MKKK_P] * y[V.MKK] / (x[C.K3] + y[V.MKK])
        v[4] = x[C.k4] * y[V.MKKK_P] * y[V.MKK_P] / (x[C.K3] + y[V.MKK_P])
        v[5] = x[C.V5] * y[V.MKK_PP] / (x[C.K5] + y[V.MKK_PP])
        v[6] = x[C.V6] * y[V.MKK_P] / (x[C.K6] + y[V.MKK_P])
        v[7] = x[C.k7] * y[V.MKK_PP] * y[V.MAPK] / (x[C.K7] + y[V.MAPK])
        v[8] = x[C.k8] * y[V.MKK_PP] * y[V.MAPK_P] / (x[C.K8] + y[V.MAPK_P])
        v[9] = x[C.V9] * y[V.MAPK_PP] / (x[C.K9] + y[V.MAPK_PP])
        v[10] = x[C.V10] * y[V.MAPK_P] / (x[C.K10] + y[V.MAPK_P])

        return v



import inspect
import math
import sys
import time
from typing import Callable, List, Literal, Optional, Tuple, Union

import numpy as np
from scipy.integrate import OdeSolver, ode, solve_ivp
from scipy.integrate._ivp.ivp import OdeResult

__all__ = ["solve_ode", "get_steady_state"]


def solve_ode(
    diffeq: Callable,
    y0: Union[list, np.ndarray],
    t: Union[range, List[int]],
    f_params: Tuple[float, ...],
    *,
    method: Union[str, OdeSolver] = "LSODA",
    vectorized: bool = False,
    options: Optional[dict] = None,
) -> Optional[OdeResult]:
    """
    Solve a system of ordinary differential equations using ``scipy.integrate.solve_ivp()``.

    Parameters
    ----------
    diffeq : callable f(t, y, *x)
        Right-hand side of the differential equation.
    y0 : array
        Initial condition on y (can be a vector).
    t : array
        A sequence of time points for which to solve for y.
    f_params : tuple
        Model parameters.
    method : str or `OdeSolver` (default: "LSODA")
        Integration method to use.
    vectorized : bool (default: :obj:`False`)
        Whether `diffeq` is implemented in a vectorized fashion.
    options : dict, optional
        Options passed to a chosen solver.

    Returns
    -------
    sol : OdeResult
        Represents the solution of ODE.
    """
    if method not in (
        available_methods := ["RK23", "RK45", "DOP853", "Radau", "BDF", "LSODA"]
    ) and not (inspect.isclass(method) and issubclass(method, OdeSolver)):
        raise ValueError(
            "`method` must be one of {} or OdeSolver class.".format(available_methods)
        )
    if options is None:
        options = {}
    options.setdefault("rtol", 1e-8)
    options.setdefault("atol", 1e-8)
    try:
        sol = solve_ivp(
            diffeq,
            (t[0], t[-1]),
            y0,
            method=method,
            t_eval=t,
            vectorized=vectorized,
            args=f_params,
            **options,
        )
        return sol if sol.success else None
    except ValueError:
        return None


def get_steady_state(
    diffeq: Callable,
    y0: list,
    f_params: tuple,
    *,
    integrator: Literal["vode", "zvode", "lsoda"] = "lsoda",
    integrator_options: Optional[dict] = None,
    dt: float = 1,
    allclose_kws: Optional[dict] = None,
    maximum_wait_time: Union[int, float] = 60.0,
) -> List[float]:
    """
    Simulate a model from given initial conditions until it reaches steady state.

    Parameters
    ----------
    diffeq : callable f(t, y, *x)
        Right-hand side of the differential equation.
    y0 : array
        Initial condition on y (can be a vector).
    f_params : tuple
        Model parameters.
    integrator : str (default: 'lsoda')
        Name of ODE integrator to use ('vode', 'zvode', or 'lsoda').
    integrator_options : dict, optional
        A dictionary of keyword arguments to supply to the integrator.
    dt : float (default: 1.0)
        The step size used to calculate the steady state.
    allclose_kws : dict, optional
        Keyword arguments to pass to ``numpy.allclose()``.
    maximum_wait_time : int or float (default: 60.0 = 1 min.)
        The longest time a user can wait for the system to reach the steady state.

    Returns
    -------
    steady_state : List[float]
        Steady state concentrations of all species.
        Return an empty list if simulation failed.
    """
    if integrator not in (availabe_integrators := ["vode", "zvode", "lsoda"]):
        raise ValueError(f"integrator must be one of {availabe_integrators}.")
    if integrator_options is None:
        integrator_options = {}
    if integrator in ["vode", "zvode"]:
        integrator_options.setdefault("method", "bdf")
        integrator_options.setdefault("with_jacobian", True)
    integrator_options.setdefault("atol", 1e-8)
    integrator_options.setdefault("rtol", 1e-8)

    if allclose_kws is None:
        allclose_kws = {}
    allclose_kws.setdefault("rtol", 1e-3)

    sol = ode(lambda t, y, f_args: diffeq(t, y, *f_args))
    sol.set_integrator(integrator, **integrator_options)
    sol.set_initial_value(y0, 0)
    sol.set_f_params(f_params)
    ys = [y0]
    start = time.time()
    while sol.successful():
        sol.integrate(sol.t + dt)
        if (
            np.iscomplex(np.real_if_close(sol.y)).any()
            or (time.time() - start) > maximum_wait_time
        ):
            return []
        elif np.allclose(sol.y, ys[-1], **allclose_kws):
            break
        else:
            ys.append(sol.y)
    steady_state = np.real(sol.y).tolist() if sol.successful() else []
    for i, val in enumerate(steady_state):
        if math.fabs(val) < sys.float_info.epsilon:
            steady_state[i] = 0.0
    return steady_state



import os
import shutil

import numpy as np

from biomass import Model, Text2Model, optimize, run_simulation

from .C import BOUNDS, CONDITIONS, EXPERIMENTAL_DATA, INDENT, NORMALIZATION, PPMEK

PATH_TO_MODEL = os.path.join(
    os.path.dirname(__file__),
    "text_files",
    "fos_model",
)


def test_preprocessing():
    if os.path.isdir(PATH_TO_MODEL):
        shutil.rmtree(PATH_TO_MODEL)


def test_model_construction():
    cfos = Text2Model(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "fos_model.txt",
        )
    )
    cfos.convert()
    assert os.path.isdir(PATH_TO_MODEL)

    with open(os.path.join(PATH_TO_MODEL, "ode.py"), mode="r") as f1:
        lines = f1.readlines()
    for line_num, line in enumerate(lines):
        if line.startswith(f"{INDENT}def diffeq(self, t, y, *x):"):
            lines[line_num - 1] = PPMEK
        elif line.startswith(f"{2*INDENT}dydt = [0] * V.NUM"):
            lines[line_num] += CONDITIONS
            break
    else:
        assert False
    with open(os.path.join(PATH_TO_MODEL, "ode.py"), mode="w") as f1:
        f1.writelines(lines)

    with open(os.path.join(PATH_TO_MODEL, "search_param.py"), mode="r") as f2:
        lines = f2.readlines()
    for line_num, line in enumerate(lines):
        if line.startswith(f"{2*INDENT}search_rgn = convert_scale("):
            lines[line_num - 1] = BOUNDS
            break
    else:
        assert False
    with open(os.path.join(PATH_TO_MODEL, "search_param.py"), mode="w") as f2:
        f2.writelines(lines)

    with open(os.path.join(PATH_TO_MODEL, "observable.py"), mode="r") as f3:
        lines = f3.readlines()
    for line_num, line in enumerate(lines):
        if line.startswith(f"{2*INDENT}self.normalization: dict = {{}}"):
            lines[line_num] += NORMALIZATION
        elif "y0 = get_steady_state(self.diffeq, y0, tuple(x))" in line:
            lines[line_num] = line.replace(
                "y0 = get_steady_state(self.diffeq, y0, tuple(x))",
                "y0 = get_steady_state(self.diffeq, y0, tuple(x), integrator='vode')",
            )
        elif "sol = solve_ode(self.diffeq, y0, self.t, tuple(x))" in line:
            lines[line_num] = line.replace(
                "sol = solve_ode(self.diffeq, y0, self.t, tuple(x))",
                "sol = solve_ode(self.diffeq, y0, self.t, tuple(x), method='BDF')",
            )
        elif line.startswith(f"{INDENT}def set_data(self)"):
            lines = lines[: line_num + 1]
            lines[line_num] = EXPERIMENTAL_DATA
            break
    else:
        assert False
    with open(os.path.join(PATH_TO_MODEL, "observable.py"), mode="w") as f3:
        f3.writelines(lines)


def test_parameter_estimation(workers: int = -1):
    model = Model("tests.test_text2model.text_files.fos_model").create()
    assert len(model.species) == 36
    assert len(model.observables) == 8
    assert len(model.problem.bounds) == 75

    param_idx = 1
    optimize(model, param_idx, optimizer_options={"maxiter": 10, "workers": workers})
    assert os.path.isdir(os.path.join(model.path, "out", f"{param_idx}"))
    assert run_simulation(model, viz_type=str(param_idx)) is None
    simulation_results = np.load(os.path.join(model.path, "simulation_data", "simulations_1.npy"))
    assert np.isfinite(simulation_results).all()
    assert len(
        [
            file
            for file in os.listdir(os.path.join(model.path, "figure", "simulation", "1"))
            if file.endswith(".pdf") or file.endswith(".png")
        ]
    ) == len(model.observables), ", ".join(
        os.listdir(os.path.join(model.path, "figure", "simulation", "1"))
    )


def test_cleanup():
    assert os.path.isdir(PATH_TO_MODEL)
    shutil.rmtree(PATH_TO_MODEL)



from dataclasses import dataclass, field
from math import isnan, log
from typing import Callable, Dict, Union

import numpy as np
from numba import njit
from scipy.integrate import simpson


@dataclass
class SignalingMetric(object):
    """
    Signaling metric used for sensitivity analysis.
    Attributes
    ----------
    quantification : Dict[str, Callable[[np.ndarray], Union[int, float]]]
        Functions to quantify signaling metrics.
    """

    quantification: Dict[str, Callable[[np.ndarray], Union[int, float]]] = field(
        default_factory=lambda: dict(
            maximum=np.max,
            minimum=np.min,
            integral=simpson,
        ),
        init=False,
    )


@njit(cache=True, fastmath=True)
def dlnyi_dlnxj(
    signaling_metric: np.ndarray,
    num_file_ids: int,
    num_perturbed_ids: int,
    num_observables: int,
    num_conditions: int,
    rate: float,
) -> np.ndarray:
    """
    Numerical computation of sensitivities using finite difference approximations
    with 1% changes in the reaction rates, parameter values or non-zero initial values.
    Parameters
    ----------
    signaling_metric : numpy array
        Signaling metric
    num_file_ids : int
        Number of ptimized parameter sets in out/
    num_perturbed_ids : int
        Number of parameters, rate equations, or non-zero initial conditions to be perturbed.
    num_observables : int
        Number of observables in observable.py
    num_conditions : int
        Number of experimental conditions.
    rate : float ~ 1
        1.01 for 1% change.
    Returns
    -------
    sensitivity_coefficients: numpy array
    """
    EPS = 2.0**-52.0
    sensitivity_coefficients = np.empty(
        (num_file_ids, num_perturbed_ids, num_observables, num_conditions),
    )
    for i in range(num_file_ids):
        for j in range(num_perturbed_ids):
            for k in range(num_observables):
                for l in range(num_conditions):
                    if isnan(signaling_metric[i, j, k, l]):
                        sensitivity_coefficients[i, j, k, l] = np.nan
                    elif (
                        abs(signaling_metric[i, -1, k, l]) < EPS
                        or abs(signaling_metric[i, j, k, l] - signaling_metric[i, -1, k, l]) < EPS
                        or (signaling_metric[i, j, k, l] / signaling_metric[i, -1, k, l]) <= 0
                    ):
                        # 1. Signaling metric before adding perturbation is zero
                        # 2. Absolute change caused by perturbation is too small
                        # 3. Antilogarithm <= 0
                        sensitivity_coefficients[i, j, k, l] = 0.0
                    else:
                        sensitivity_coefficients[i, j, k, l] = log(
                            signaling_metric[i, j, k, l] / signaling_metric[i, -1, k, l]
                        ) / log(rate)

    return sensitivity_coefficients


def remove_nan(sensitivity_matrix: np.ndarray) -> np.ndarray:
    """
    Remove NaN from sensitivity matrix. This function is used for preprocessing of visualizing
    the result of sensitivity analysis through a heatmap.
    Parameters
    ----------
    sensitivity_matrix : numpy.ndarray
        M x N matrix, where M and M are # of parameter sets and # of perturbed objects, respectively.
    """
    EPS = 2.0**-52.0
    nan_idx = []
    for i in range(sensitivity_matrix.shape[0]):
        if np.isnan(sensitivity_matrix[i, :]).any():
            nan_idx.append(i)
        if np.nanmax(np.abs(sensitivity_matrix[i, :])) < EPS:
            sensitivity_matrix[i, :] = np.zeros(sensitivity_matrix.shape[1])
    return np.delete(sensitivity_matrix, nan_idx, axis=0)



"""
`biomass.models` contains sample models and other scripts if you would prefer to learn how to use biomass by example.
For details, please refer to https://biomass-core.readthedocs.io/en/latest/models.html.
"""

import os
import shutil
from typing import Literal


def copy_to_current(
    model_name: Literal[
        "circadian_clock",
        "insulin_signaling",
        "mapk_cascade",
        "Nakakuki_Cell_2010",
        "nfkb_pathway",
        "pan_rtk",
        "tgfb_smad",
        "g1s_transition",
        "prolif_quies",
    ]
) -> None:
    """Copy an example model to the current working directory.

    To execute example models in your machine, please follow these steps:

    >>> from biomass import create_model
    >>> from biomass.models import copy_to_current
    >>> copy_to_current('XXX')
    >>> model = create_model('XXX')

    Parameters
    ----------
    model_name : str
        Name of the example model.
    """
    if model_name not in (
        available_models := [
            "circadian_clock",
            "insulin_signaling",
            "mapk_cascade",
            "Nakakuki_Cell_2010",
            "nfkb_pathway",
            "pan_rtk",
            "tgfb_smad",
            "g1s_transition",
            "prolif_quies",
        ]
    ):
        raise ValueError(f"model_name must be one of [{', '.join(available_models)}].")
    this_directory = os.path.abspath(os.path.dirname(__file__))
    shutil.copytree(
        os.path.join(this_directory, model_name), os.path.join(os.getcwd(), model_name)
    )



import os
import shutil
from typing import Callable

import numpy as np
import pytest

from biomass import Model, Text2Model, run_simulation
from biomass.construction.thermodynamic_restrictions import DuplicateError


def test_preprocessing():
    for model in [
        "michaelis_menten",
        "Kholodenko1999",
        "michaelis_menten_jl",
        "Kholodenko1999_jl",
    ]:
        if os.path.isdir(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                model,
            )
        ):
            shutil.rmtree(
                os.path.join(
                    os.path.dirname(__file__),
                    "text_files",
                    model,
                )
            )


def test_text2model():
    for model in ["michaelis_menten", "Kholodenko1999"]:
        for lang in ["python", "julia"]:
            if model == "michaelis_menten":
                mm_kinetics = Text2Model(
                    os.path.join(
                        os.path.dirname(__file__),
                        "text_files",
                        f"{model}.txt",
                    ),
                    similarity_threshold=0.7,
                    lang=lang,
                )
                mm_kinetics.register_word({"dissociate": ["releases"]})
                mm_kinetics.convert()
            elif model == "Kholodenko1999":
                mapk_cascade = Text2Model(
                    os.path.join(
                        os.path.dirname(__file__),
                        "text_files",
                        f"{model}.txt",
                    ),
                    lang=lang,
                )
                mapk_cascade.convert()
                # test thermodynamic restrictions
                desired = [
                    {"10", "11", "12", "9"},
                    {"15", "17", "18", "21"},
                    {"18", "19", "20", "22"},
                    {"18", "19", "20", "22"},
                    {"12", "17", "19", "24"},
                    {"15", "20", "23", "24"},
                    {"12", "21", "22", "23"},
                ]
                actual = [set(restriction) for restriction in mapk_cascade.restrictions]
                for l1 in actual:
                    assert l1 in desired
                for l2 in desired:
                    assert l2 in actual

    with pytest.raises(DuplicateError):
        Text2Model(
            os.path.join(os.path.dirname(__file__), "text_files", "duplicate_binding.txt")
        ).convert()


def test_run_simulation():
    try:
        from .text_files import Kholodenko1999, michaelis_menten

        _packaging: Callable[[str], str] = lambda model_name: ".".join(
            ["tests.test_text2model.text_files", model_name]
        )
        for model_name in ["Kholodenko1999", "michaelis_menten"]:
            model = Model(_packaging(model_name)).create()
            run_simulation(model)
            simulated_values = np.load(
                os.path.join(
                    os.path.dirname(__file__),
                    "text_files",
                    model_name,
                    "simulation_data",
                    "simulations_original.npy",
                )
            )
            assert np.isfinite(simulated_values).all()
            if model_name == "Kholodenko1999":
                expected = np.load(
                    os.path.join(
                        os.path.dirname(__file__),
                        "simulations_original_BN.npy",
                    )
                )
                assert np.allclose(simulated_values, expected)
    except ImportError as e:
        print(e)


def test_text2markdown():
    for model in ["michaelis_menten", "Kholodenko1999"]:
        if model == "michaelis_menten":
            mm_kinetics = Text2Model(
                os.path.join(
                    os.path.dirname(__file__),
                    "text_files",
                    f"{model}.txt",
                )
            )
            mm_kinetics.register_word({"dissociate": ["releases"]})
            mm_kinetics.to_markdown(num_reactions=2)
        elif model == "Kholodenko1999":
            mapk_cascade = Text2Model(
                os.path.join(
                    os.path.dirname(__file__),
                    "text_files",
                    f"{model}.txt",
                )
            )
            mapk_cascade.to_markdown(num_reactions=25)
        assert os.path.isfile(os.path.join("markdown", model, "rate.md"))
        assert os.path.isfile(os.path.join("markdown", model, "diffeq.md"))
        shutil.rmtree("markdown")


def test_julia_models():
    necessities = [
        os.path.join("name2idx", "parameters.jl"),
        os.path.join("name2idx", "species.jl"),
        "ode.jl",
        "observable.jl",
        "simulation.jl",
        "experimental_data.jl",
        "search_param.jl",
        "problem.jl",
    ]
    for model in ["michaelis_menten", "Kholodenko1999"]:
        for file in necessities:
            assert os.path.isfile(
                os.path.join(
                    os.path.dirname(__file__),
                    "text_files",
                    f"{model}_jl",
                    file,
                )
            )


def test_cleanup():
    for model in [
        "michaelis_menten",
        "Kholodenko1999",
        "michaelis_menten_jl",
        "Kholodenko1999_jl",
        "duplicate_binding",
    ]:
        assert os.path.isdir(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                model,
            )
        )
        shutil.rmtree(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                model,
            )
        )



import os
import shutil

import numpy as np

from biomass import OptimizationResults, create_model, optimize, run_analysis, run_simulation
from biomass.models import copy_to_current

MODEL_NAME: str = "mapk_cascade"

copy_to_current(MODEL_NAME)
assert os.path.exists(MODEL_NAME)
model = create_model(MODEL_NAME)


def test_simulate_successful():
    x = model.pval()
    y0 = model.ival()
    assert model.problem.simulate(x, y0) is None


def test_optimize():
    for x_id in range(1, 4):
        optimize(model, x_id=x_id, optimizer_options={"maxiter": 5, "workers": -1})
    for paramset in range(1, 4):
        with open(
            os.path.join(
                model.path,
                "out",
                f"{paramset:d}",
                "optimization.log",
            )
        ) as f:
            logs = f.readlines()
        assert logs[-1].startswith("differential_evolution step 5: ")


def test_run_simulation():
    run_simulation(model, viz_type="original")
    run_simulation(model, viz_type="1")
    run_simulation(model, viz_type="2")
    run_simulation(model, viz_type="3")
    run_simulation(model, viz_type="average", stdev=True)
    run_simulation(model, viz_type="best", show_all=True)
    for npy_file in [
        "simulations_original.npy",
        "simulations_1.npy",
        "simulations_2.npy",
        "simulations_3.npy",
        "simulations_all.npy",
        "simulations_best.npy",
    ]:
        simulated_value = np.load(os.path.join(model.path, "simulation_data", npy_file))
        assert np.isfinite(simulated_value).all()
    for viz_type in ["original", "1", "2", "3", "average", "best"]:
        for obs_name in model.observables:
            assert os.path.isfile(
                os.path.join(model.path, "figure", "simulation", f"{viz_type}", f"{obs_name}.png")
            )


def test_save_result():
    res = OptimizationResults(model)
    res.to_csv()
    assert os.path.isfile(
        os.path.join(
            model.path,
            "optimization_results",
            "optimized_params.csv",
        )
    )
    res.savefig()
    assert os.path.isfile(
        os.path.join(
            model.path,
            "optimization_results",
            "estimated_parameter_sets.pdf",
        )
    )
    res.dynamic_assessment(include_original=True)
    assert os.path.isfile(
        os.path.join(
            model.path,
            "optimization_results",
            "fitness_assessment.csv",
        )
    )
    res.trace_obj()
    assert os.path.isfile(
        os.path.join(
            model.path,
            "optimization_results",
            "obj_func_traces.pdf",
        )
    )


def _get_duration(
    time_course: np.ndarray,
    below_threshold: float = 0.5,
) -> int:
    """
    Calculation of the duration as the time it takes to decline below the threshold.

    Parameters
    ----------
    timecourse : array
        Simulated time course data.

    below_threshold : float (from 0.0 to 1.0)
        0.5 for 50% of its maximum.

    Returns
    -------
    duration : int

    """
    if not 0.0 < below_threshold < 1.0:
        raise ValueError("below_threshold must lie within (0.0, 1.0).")
    maximum_value = np.max(time_course)
    t_max = np.argmax(time_course)
    time_course = time_course - below_threshold * maximum_value
    time_course[time_course > 0.0] = -np.inf
    duration = np.argmax(time_course[t_max:]) + t_max

    return duration


def test_sensitivity_analysis():
    for target in [
        # "parameter",
        "initial_condition",
        # "reaction",
    ]:
        for metric in [
            "maximum",
            "minimum",
            "integral",
            "argmax",
            "timepoint",
            "duration",
        ]:
            run_analysis(
                model,
                target=target,
                metric=metric,
                create_metrics={
                    "argmax": np.argmax,
                    "timepoint": lambda time_course: time_course[model.problem.t[-1]],
                    "duration": _get_duration,
                },
            )
            sensitivity_coefficients = np.load(
                os.path.join(
                    model.path,
                    "sensitivity_coefficients",
                    target,
                    f"{metric}.npy",
                )
            )
            assert np.isfinite(sensitivity_coefficients).all()


def test_cleanup():
    shutil.rmtree(MODEL_NAME)



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "Bg",
    "kSyE2f",
    "kSyE2fE2f",
    "jSyE2f",
    "kAsRbE2f",
    "kDsRbE2f",
    "kDeE2f",
    "kPhRbCd",
    "kPhRbCe",
    "kPhRbCa",
    "kDpRb",
    "kSyE1",
    "kDeE1C1",
    "kDeE1",
    "kPhC1",
    "kPhC1Ce",
    "kPhC1Ca",
    "kDpC1",
    "kAsE1C1",
    "kDsE1C1",
    "kSyP21",
    "kSyP21P53",
    "kDeP21",
    "kDeP21Cy",
    "kDeP21aRc",
    "kSyCe",
    "kSyCa",
    "kAsCyP21",
    "kDsCyP21",
    "kDeCe",
    "kDeCa",
    "kDeCeCa",
    "kDeCaC1",
    "kImPc",
    "kExPc",
    "kPhRc",
    "kDpRc",
    "jCy",
    "n",
    "kAsRcPc",
    "kDsRcPc",
    "kAsPcP21",
    "kDsPcP21",
    "kSyDna",
    "kSyP53",
    "kDeP53",
    "jP53",
    "kGeDam",
    "kGeDamArc",
    "kReDam",
    "kReDamP53",
    "jDam",
    "kSyPr",
    "kDePr",
]

NUM: int = len(NAMES)

Parameters = make_dataclass(
    cls_name="Parameters",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

C = Parameters(**name2idx)

del name2idx



"""BioMASS core functions"""
import os
import warnings
from dataclasses import dataclass
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import Callable, Dict, Literal, Optional, Union

import numpy as np
from scipy.optimize import differential_evolution

from .analysis import InitialConditionSensitivity, ParameterSensitivity, ReactionSensitivity
from .dynamics import SignalingSystems
from .estimation import Optimizer
from .model_object import ModelObject
from .version import __version__

__all__ = ["Model", "create_model", "optimize", "run_simulation", "run_analysis"]


class BiomassIndexError(Exception):
    """
    Error in specifying model paramters or species.
    """

    pass


@dataclass
class Model(object):
    """
    Class for BioMASS model construction.

    Attributes
    ----------
    pkg_name: str
        Path (dot-sepalated) to a biomass model directory.
        Use ``__package__``.
    """

    pkg_name: str

    def _load_model(self) -> Optional[ModuleType]:
        try:
            biomass_model = import_module(self.pkg_name)
            return biomass_model
        except ImportError:
            p = Path(self.pkg_name.replace(".", os.sep))
            print(f"cannot import '{p.name}' from '{p.parent}'.")
        return None

    @staticmethod
    def _check_indices(model: ModelObject) -> None:
        """
        Check indices used in a BioMASS model.
        'C' and 'V' must be used for parameters and species, respectively.
        """
        files = ["ode.py", "search_param.py", "observable.py", "reaction_network.py"]
        for file in files:
            with open(os.path.join(model.path, file)) as f:
                lines = f.readlines()
            for i, line in enumerate(lines, start=1):
                loc = f"line {i:d} in {file}"
                msg = ", Use '{}' for the index of a {}."
                if "x[V." in line:
                    raise BiomassIndexError(loc + msg.format("C", "parameter"))
                elif any(map(line.__contains__, ("y[C.", "y0[C.", "dydt[C."))):
                    raise BiomassIndexError(loc + msg.format("V", "species"))

    @staticmethod
    def _check_normalization(model: ModelObject) -> None:
        """
        Check normalization options in ``observable.py``.
        """
        if model.problem.normalization:
            for obs_name in model.observables:
                for key in model.problem.normalization[obs_name]:
                    if key not in ["timepoint", "condition"]:
                        raise KeyError(
                            f"'{key}': invalid key. Should be either 'timepoint' or 'condition'."
                        )
                if (
                    isinstance(model.problem.normalization[obs_name]["timepoint"], int)
                    and not model.problem.t[0]
                    <= model.problem.normalization[obs_name]["timepoint"]
                    <= model.problem.t[-1]
                ):
                    raise ValueError("Normalization timepoint must lie within problem.t.")
                if not model.problem.normalization[obs_name]["condition"]:
                    model.problem.normalization[obs_name]["condition"] = model.problem.conditions
                else:
                    for c in model.problem.normalization[obs_name]["condition"]:
                        if c not in model.problem.conditions:
                            raise ValueError(
                                f"Normalization condition '{c}' is not defined in problem.conditions."
                            )

    @staticmethod
    def _check_visualization_options(model: ModelObject) -> None:
        """
        Check visualization options in ``viz.py``.
        """
        err_msg = (
            "{} in viz.py, len({}) must be equal to or greater than "
            f"len(self.model.problem.conditions) (={len(model.problem.conditions)})."
        )
        singleplotting = model.viz.get_single_observable_options()
        multiplotting = model.viz.get_multiple_observables_options()
        sensitivity_options = model.viz.get_sensitivity_options()
        # Visualization options for time-course simulation (single-observable)
        for i, obs_name in enumerate(model.observables):
            if len(singleplotting[i].cmap) < len(model.problem.conditions):
                raise ValueError(
                    err_msg.format(
                        f"self.single_observable_options for {obs_name}",
                        f"self.single_observable_options[{i}].cmap",
                    )
                )
            elif model.problem.experiments[i] is not None and len(singleplotting[i].shape) < len(
                model.problem.conditions
            ):
                raise ValueError(
                    err_msg.format(
                        f"self.single_observable_options for {obs_name}",
                        f"self.single_observable_options[{i}].shape",
                    )
                )
        # Visualization options for time-course simulation (multi-observables)
        if len(multiplotting.cmap) < len(model.problem.conditions):
            raise ValueError(
                err_msg.format(
                    "self.multiple_observables_options",
                    "self.multiple_observables_options.cmap",
                )
            )
        for obs_name in multiplotting.observables:
            if model.problem.experiments[model.observables.index(obs_name)] is not None and len(
                multiplotting.shape
            ) < len(model.problem.conditions):
                raise ValueError(
                    err_msg.format(
                        "self.multiple_observables_options",
                        "self.multiple_observables_options.shape",
                    )
                )
        # Visualization options for sensitivity analysis results
        if len(sensitivity_options.cmap) < len(model.problem.conditions):
            raise ValueError(
                err_msg.format(
                    "self.sensitivity_options",
                    "self.sensitivity_options.cmap",
                )
            )

    def create(self, show_info: bool = False) -> ModelObject:
        """
        Build a biomass model.

        Parameters
        ----------
        show_info : bool (default: :obj:`False`)
            Set to :obj:`True` to print the information related to model size.

        Returns
        -------
        model : :class:`biomass.model_object.ModelObject`
            The BioMASS model object.

        Examples
        --------
        >>> from biomass import Model
        >>> import your_model
        >>> model = Model(your_model.__package__).create()
        """
        model = ModelObject(self.pkg_name.replace(".", os.sep), self._load_model())
        self._check_indices(model)
        self._check_normalization(model)
        self._check_visualization_options(model)
        if show_info:
            model_name = Path(model.path).name
            print(f"{model_name} information\n" + ("-" * len(model_name)) + "------------")
            print(f"{len(model.species):d} species")
            print(
                f"{len(model.parameters):d} parameters, "
                f"of which {len(model.problem.idx_params):d} to be estimated"
            )
        return model


def create_model(pkg_name: str, show_info: bool = False) -> ModelObject:
    """
    Create a BioMASS model.

    Parameters
    ----------
    pkg_name : str
        Path (dot-sepalated) to a biomass model directory.
    show_info : bool (default: :obj:`False`)
        Set to :obj:`True` to print the information related to model size.

    Returns
    -------
    model : :class:`biomass.model_object.ModelObject`
        The BioMASS model object.

    Examples
    --------
    >>> from biomass import create_model
    >>> import your_model
    >>> model = create_model(your_model.__package__)
    """
    model = Model(pkg_name).create(show_info)
    return model


def optimize(
    model: ModelObject,
    x_id: int,
    *,
    disp_here: bool = False,
    overwrite: bool = False,
    optimizer_options: Optional[dict] = None,
) -> None:
    """
    Estimate model parameters from experimental data.

    Parameters
    ----------
    model : ModelObject
        Model for parameter estimation.
    x_id : int
        Index of parameter set to estimate.
    disp_here : bool (default: :obj:`False`)
        Whether to show the evaluated *objective* at every iteration.
    overwrite : bool (default: :obj:`False`)
        If :obj:`True`, the directory (``x_id/``) will be overwritten.
    optimizer_options : dict, optional
        Keyword arguments to pass to ``scipy.optimize.differential_evolution``.
        For details, please refer to https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.differential_evolution.html.

    Examples
    --------
    >>> from biomass import create_model, optimize
    >>> from biomass.models import copy_to_current
    >>> copy_to_current("Nakakuki_Cell_2010")
    >>> model = create_model("Nakakuki_Cell_2010")
    >>> optimize(model, x_id=1)

    Notes
    -----

    * Set simulation conditions and the corresponding experimental data in ``observable.py``
    * Define an objective function to be minimized (:func:`objective`) in ``problem.py``
    * Set lower/upper bounds of parameters to be estimated in ``search_param.py``

    """
    if optimizer_options is None:
        optimizer_options = {}
    optimizer_options.setdefault("strategy", "best1bin")
    optimizer_options.setdefault("maxiter", 50)
    optimizer_options.setdefault("popsize", 3)
    optimizer_options.setdefault("tol", 1e-4)
    optimizer_options.setdefault("mutation", 0.1)
    optimizer_options.setdefault("recombination", 0.5)
    optimizer_options.setdefault("disp", True)
    optimizer_options.setdefault("polish", False)
    optimizer_options.setdefault("workers", 1)

    if not optimizer_options["disp"]:
        raise ValueError(
            "Set optimizer_options['disp'] to True. "
            "If you don't want to see the evaluated objective function at every iteration, "
            "set the keyword argument `disp_here` to False."
        )
    if optimizer_options["polish"]:
        warnings.warn(
            "Setting optimizer_options['polish'] to False is highly recommended.",
            UserWarning,
        )

    optimizer = Optimizer(model, differential_evolution, x_id, disp_here, overwrite)
    res = optimizer.minimize(
        model.get_obj_val,
        [(0, 1) for _ in range(len(model.problem.bounds))],
        **optimizer_options,
    )
    param_values = model.gene2val(res.x)
    optimizer.import_solution(param_values)


def run_simulation(
    model: ModelObject,
    *,
    viz_type: str = "original",
    show_all: bool = False,
    stdev: bool = False,
) -> None:
    """
    Simulate ODE model with estimated parameter values.

    Parameters
    ----------
    model : ModelObject
        Model for simulation.
    viz_type : str
        * 'average':
            The average of simulation results with parameter sets in ``out/``.
        * 'best':
            The best simulation result in ``out/``, simulation with `best_fit_param`.
        * 'original':
            Simulation with the default parameters and initial values defined in ``set_model.py``.
        * 'n(=1,2,...)':
            Use the parameter set in ``out/n/``.
        * 'experiment'
            Draw the experimental data written in ``observable.py`` without simulation results.
    show_all : bool (default: :obj:`False`)
        Whether to show all simulation results.
    stdev : bool (default: :obj:`False`)
        If :obj:`True`, the standard deviation of simulated values will be shown
        (only available for 'average' visualization type).

    Examples
    --------
    >>> from biomass import create_model, run_simulation
    >>> from biomass.models import copy_to_current
    >>> copy_to_current("Nakakuki_Cell_2010")
    >>> model = create_model("Nakakuki_Cell_2010")
    >>> run_simulation(
    ...     model,
    ...     viz_type='average',
    ...     show_all=False,
    ...     stdev=True,
    ... )

    """
    if viz_type not in ["best", "average", "original", "experiment"] and not viz_type.isdecimal():
        raise ValueError(
            "Available viz_type are: 'best','average','original','experiment','n(=1, 2, ...)'"
        )

    SignalingSystems(model).simulate_all(
        viz_type=viz_type,
        show_all=show_all,
        stdev=stdev,
    )


def run_analysis(
    model: ModelObject,
    *,
    target: Literal["reaction", "parameter", "initial_condition"],
    metric: str = "integral",
    create_metrics: Optional[Dict[str, Callable[[np.ndarray], Union[int, float]]]] = None,
    style: Literal["barplot", "heatmap"] = "barplot",
    show_progress: bool = True,
    clustermap_kws: Optional[dict] = None,
    cbar_ax_tick_params: Optional[dict] = None,
    options: Optional[dict] = None,
) -> None:
    """
    Employ sensitivity analysis to identify critical parameters, species or
    reactions in the complex biological network.

    The sensitivity S(y,x) was calculated according to the following equation:
    S(y,x) = d ln(yi) / d ln (xj), where yi is the signaling metric and xj is
    each nonzero species, parameter value or reaction rate.

    Parameters
    ----------
    model : ModelObject
        Model for sensitivity analysis.
    target : Literal["reaction", "parameter", "initial_condition"]
        Where to add a small perturbation to calculate sensitivity coefficients.
    metric : str (default: 'integral')
        A word to specify the signaling metric.
    create_metrics : Dict[str, Callable[[np.ndarray], Union[int, float]]], optional
        Create user-defined signaling metrics.
    style :  Literal["barplot", "heatmap"] (default: 'barplot')
        * 'barplot'
        * 'heatmap'
    show_progress : bool (default: :obj:`True`)
        Set to :obj:`True` to show the progress indicator while calculating sensitivity coefficients.
    clustermap_kws : dict, optional
        Keyword arguments to pass to ``seaborn.clustermap()`` when style is 'heatmap'.
    cbar_ax_tick_params : dict, optional
        Keyword arguments to pass to ``ax.tick_params()`` of the color bar object when style is 'heatmap'.

    options : dict, optional
        * show_indices : bool (default: :obj:`True`)
            (``target`` == 'reaction') Set to :obj:`True` to put reaction index on each bar.

        * excluded_params : list of strings
            (``target`` == 'parameter') List of parameters which are not used for analysis.

        * excluded_initials : list of strings
            (``target`` == 'initial_condition') List of species which are not used for analysis.

        * overwrite : bool (default: :obj:`True`)
            If :obj:`True`, the sensitivity_coefficients/{target}/{metric}.npy file will be overwritten.

    Examples
    --------
    >>> from biomass import create_model, run_analysis
    >>> from biomass.models import copy_to_current
    >>> copy_to_current("Nakakuki_Cell_2010")
    >>> model = create_model("Nakakuki_Cell_2010")

    Parameters

    >>> run_analysis(
    ...     model,
    ...     target='parameter',
    ...     options = {
    ...         'excluded_params': [
    ...             'a', 'Vn', 'Vc', 'Ligand', 'EGF', 'HRG', 'no_ligand'
    ...         ]
    ...     }
    ... )

    Initial condition

    >>> run_analysis(model, target='initial_condition')

    Reaction

    >>> run_analysis(model, target='reaction')

    """
    if clustermap_kws is None:
        clustermap_kws = {}
    clustermap_kws.setdefault("figsize", (12, 5))
    clustermap_kws.setdefault("center", 0)
    clustermap_kws.setdefault("robust", True)
    clustermap_kws.setdefault("method", "ward")
    clustermap_kws.setdefault("cmap", "RdBu_r")
    clustermap_kws.setdefault("linewidth", 0.5)
    clustermap_kws.setdefault("col_cluster", False)
    clustermap_kws.setdefault("yticklabels", [])

    if cbar_ax_tick_params is None:
        cbar_ax_tick_params = {}
    cbar_ax_tick_params.setdefault("labelsize", 8)

    if options is None:
        options = {}
    options.setdefault("show_indices", True)
    options.setdefault("excluded_params", [])
    options.setdefault("excluded_initials", [])
    options.setdefault("overwrite", True)

    if target == "reaction":
        ReactionSensitivity(model, create_metrics).analyze(
            metric=metric,
            style=style,
            show_progress=show_progress,
            clustermap_kws=clustermap_kws,
            cbar_ax_tick_params=cbar_ax_tick_params,
            options=options,
        )
    elif target == "parameter":
        ParameterSensitivity(model, create_metrics).analyze(
            metric=metric,
            style=style,
            show_progress=show_progress,
            clustermap_kws=clustermap_kws,
            cbar_ax_tick_params=cbar_ax_tick_params,
            options=options,
        )
    elif target == "initial_condition":
        InitialConditionSensitivity(model, create_metrics).analyze(
            metric=metric,
            style=style,
            show_progress=show_progress,
            clustermap_kws=clustermap_kws,
            cbar_ax_tick_params=cbar_ax_tick_params,
            options=options,
        )
    else:
        raise ValueError(
            "Available targets are: '{}".format(
                "', '".join(["reaction", "parameter", "initial_condition"]) + "'."
            )
        )



from typing import Dict, List


class ReactionNetwork(object):
    """
    Reaction indices grouped according to biological processes.
    This is used for sensitivity analysis (target='reaction').
    """

    def __init__(self) -> None:
        self.reactions: Dict[str, List[int]] = {}



import os
import shutil

import matplotlib.pyplot as plt
import numpy as np

from biomass import create_model, run_simulation
from biomass.models import copy_to_current

MODEL_NAME: str = "prolif_quies"

copy_to_current(MODEL_NAME)
assert os.path.exists(MODEL_NAME)
model = create_model(MODEL_NAME)


def test_simulate_successful():
    x = model.pval()
    y0 = model.ival()
    assert model.problem.simulate(x, y0) is None


def test_run_simulation():
    assert run_simulation(model) is None
    res = np.load(os.path.join(model.path, "simulation_data", "simulations_original.npy"))

    plt.figure(figsize=(10, 5))
    plt.rcParams["font.size"] = 32
    plt.rcParams["axes.linewidth"] = 1.5
    plt.rcParams["lines.linewidth"] = 6

    plt.plot(
        [t_ / 60 for t_ in model.problem.t],
        res[model.observables.index("CycA"), 0],
        color="slateblue",
        label="CycA",
    )
    plt.plot(
        [t_ / 60 for t_ in model.problem.t],
        res[model.observables.index("CycE"), 0],
        color="skyblue",
        label="CycE",
    )
    plt.plot(
        [t_ / 60 for t_ in model.problem.t],
        res[model.observables.index("active_RC"), 0],
        color="firebrick",
        label="aRC",
    )
    plt.plot(
        [t_ / 60 for t_ in model.problem.t],
        res[model.observables.index("p21_tot"), 0],
        color="limegreen",
        label="p21",
    )

    plt.xlim(0, 20)
    plt.xlabel("time from cytokinesis (h)")
    plt.ylim(0, 2)
    plt.ylabel("rel. level (AU)")
    plt.xticks([0, 5, 10, 15, 20])
    plt.yticks([0, 1, 2])
    plt.legend(loc="upper left", frameon=False, fontsize=18)

    plt.show()


def test_cleanup():
    shutil.rmtree(MODEL_NAME)



from collections import Counter
from dataclasses import dataclass, field
from typing import List, NamedTuple, Tuple


class DuplicateError(Exception):
    pass


class ComplexFormation(NamedTuple):
    rxn_no: int
    components: set
    complex: str
    is_binding: bool


@dataclass
class ThermodynamicRestrictions(object):
    """
    Thermodynamic restrictions along cyclic pathways in a kinetic scheme.

    Notes
    -----
    If a kinetic scheme includes "true" cycles, in which the initial and final states are
    identical, the equilibrium constants of the reactions along any cycle satisfy so-called
    "detailed balance" relationships. These detailed balance relations require the product
    of the equilibrium constants along a cycle to be equal to 1, since at equilibrium the net flux
    through any cycle vanishes.
    """

    _rxn_indices: dict = field(
        default_factory=dict,
        init=False,
    )
    _tree: dict = field(
        default_factory=dict,
        init=False,
    )
    restrictions: list = field(
        default_factory=list,
        init=False,
    )
    complex_formations: List[ComplexFormation] = field(
        default_factory=list,
        init=False,
    )

    def _create_tree(
        self,
        tree: dict,
        parent_pattern: ComplexFormation,
    ) -> dict:
        """
        Create tree.
        """
        for component in parent_pattern.components:
            for another_pattern in self.complex_formations:
                if (
                    component == another_pattern.complex
                    and parent_pattern.rxn_no != another_pattern.rxn_no
                ):
                    for k in another_pattern.components:
                        if component not in tree.keys():
                            tree.setdefault(component, {})
                        tree[component].setdefault(k, {f"{another_pattern.rxn_no}": {}})
                        tree[component][k] = self._create_tree(tree[component][k], another_pattern)
        return tree

    def _add_to_rxn_indices1(self, complex_name: str, tree: dict) -> None:
        for component in tree.keys():
            if not tree[component]:
                self._rxn_indices[complex_name].append(component)
            else:
                self._add_to_rxn_indices1(complex_name, tree[component])

    def _add_to_rxn_indices2(self, complex_name: str, monomer: str, pair: str) -> None:
        for another_complex in self._tree[complex_name].keys():
            if (
                monomer in self._tree[complex_name][another_complex].keys()
                and len(self._tree[complex_name][another_complex][monomer]) > 1
            ):
                for name in self._tree[complex_name][another_complex][monomer].keys():
                    if name in self._tree[complex_name][pair].keys():
                        self._append_reaction_number(
                            complex_name,
                            monomer,
                            pair,
                            name,
                            another_complex,
                        )

    def _add_to_rxn_indices3(self, complex_name: str, monomer: str, pair: str) -> None:
        for another_complex in self._tree[complex_name].keys():
            if (
                monomer in self._tree[complex_name][another_complex].keys()
                and len(self._tree[complex_name][another_complex][monomer]) == 1
            ):
                for name in self._tree[complex_name][another_complex].keys():
                    if not name.isdecimal() and name != monomer:
                        for components in self._tree[complex_name][another_complex].keys():
                            if name == components:
                                for k in self._tree[complex_name][pair][name].keys():
                                    if not k.isdecimal() and k in self._tree[complex_name].keys():
                                        self._append_reaction_number(
                                            complex_name,
                                            monomer,
                                            pair,
                                            name,
                                            another_complex,
                                        )

    def _append_reaction_number(
        self,
        complex_name: str,
        monomer: str,
        pair: str,
        name: str,
        another_complex: str,
    ) -> None:
        for val in self._tree[complex_name][another_complex].keys():
            if val.isdecimal():
                self._rxn_indices[monomer].append(val)
        for val in self._tree[complex_name][another_complex][monomer].keys():
            if val.isdecimal():
                self._rxn_indices[monomer].append(val)
        for val in self._tree[complex_name][pair][name].keys():
            if val.isdecimal():
                self._rxn_indices[monomer].append(val)

    def _get_complex_patterns(self) -> List[Tuple[ComplexFormation, ComplexFormation]]:
        complex_patterns = []
        for i, pattern_a in enumerate(self.complex_formations):
            for j in range(i + 1, len(self.complex_formations)):
                pattern_b = self.complex_formations[j]
                if (
                    pattern_a.components == pattern_b.components
                    and pattern_a.complex == pattern_b.complex
                ):
                    raise DuplicateError(
                        "Duplicate binding-dissociation events are detected "
                        f"in lines {pattern_a.rxn_no:d} and {pattern_b.rxn_no:d}."
                    )
                elif (
                    pattern_a.is_binding != pattern_b.is_binding
                    and pattern_a.complex == pattern_b.complex
                ):
                    complex_patterns.append((pattern_a, pattern_b))
        return complex_patterns

    def _impose_restrictions(self, complex_name: str) -> None:
        count = Counter(self._rxn_indices[complex_name])
        if all(n == 2 for n in count.values()):
            self.restrictions.append(list(set(self._rxn_indices[complex_name])))
        else:
            _monomers = {}
            for name in self._tree[complex_name].keys():
                if len(self._tree[complex_name][name].values()) == 1:
                    _monomers[name] = list(self._tree[complex_name][name].keys())[0]
            for monomer, ridx in _monomers.items():
                self._rxn_indices[monomer] = [ridx]
                pair = None
                for another_complex in self._tree[complex_name].keys():
                    if (
                        another_complex != monomer
                        and ridx in self._tree[complex_name][another_complex]
                    ):
                        pair = another_complex
                if pair is not None:
                    self._add_to_rxn_indices2(complex_name, monomer, pair)
                    _reactions = list(set(self._rxn_indices[monomer]))
                    if len(_reactions) > 2 and _reactions not in self.restrictions:
                        self.restrictions.append(_reactions)
            # initialize monomers
            for monomer, ridx in _monomers.items():
                self._rxn_indices[monomer] = [ridx]
                pair = None
                for another_complex in self._tree[complex_name].keys():
                    if (
                        another_complex != monomer
                        and ridx in self._tree[complex_name][another_complex]
                    ):
                        pair = another_complex
                if pair is not None:
                    self._add_to_rxn_indices3(complex_name, monomer, pair)
                    _reactions = list(set(self._rxn_indices[monomer]))
                    if len(_reactions) > 2 and _reactions not in self.restrictions:
                        self.restrictions.append(_reactions)

    def find_cyclic_reaction_routes(self) -> None:
        """
        Find cyclic pathways in a reaction network.
        """
        complex_patterns = self._get_complex_patterns()
        _tree = {}
        for patterns in complex_patterns:
            for pattern in patterns:
                _tree.setdefault(pattern.complex, {})
                for component in pattern.components:
                    _tree[pattern.complex].setdefault(component, {f"{pattern.rxn_no}": {}})
                _tree[pattern.complex] = self._create_tree(_tree[pattern.complex], pattern)
        self._tree = _tree
        for complex_name in self._tree.keys():
            self._rxn_indices.setdefault(complex_name, [])
            self._add_to_rxn_indices1(complex_name, self._tree[complex_name])
            self._impose_restrictions(complex_name)



from typing import Dict, List

from .name2idx import C, V


class ReactionNetwork(object):
    def __init__(self) -> None:
        """
        Reaction indices grouped according to biological processes.
        This is used for sensitivity analysis (target='reaction').
        """
        super(ReactionNetwork, self).__init__()

        self.reactions: Dict[str, List[int]] = {
            "all": [i for i in range(1, 114)],
        }

    @staticmethod
    def flux(t, y, x) -> dict:
        """
        Rate equations in the model.

        Parameters
        ----------
        t : float
            Time point.
        y : ndarray
            Concentration vector.
        x : tuple
            Parameter values.

        Returns
        -------
        v : dict
            Flux vector.
        """
        v = {}

        return v



import os
import shutil

from biomass import Text2Model

file_dir = os.path.join(os.path.dirname(__file__), "text_files")
txt_files = [name for name in os.listdir(file_dir) if "__" not in name]

skipped_files = ["duplicate_binding.txt", "typo_1.txt", "typo_2.txt", "unregistered_rule.txt"]


def test_preprocessing():
    for model in txt_files:
        model = model.split(".")[0]
        if os.path.isdir(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                model,
            )
        ):
            shutil.rmtree(
                os.path.join(
                    os.path.dirname(__file__),
                    "text_files",
                    model,
                )
            )


def test_graph():
    for model_file in txt_files:
        if model_file in skipped_files:
            continue
        model = Text2Model(os.path.join(os.path.dirname(__file__), "text_files", model_file))
        model_path = os.path.join(file_dir, model_file.split(".")[0])
        model.register_word({"dissociate": ["releases"]})
        model.convert(overwrite=True)
        model.static_plot(save_dir=model_path, file_name="test.png")
        model.dynamic_plot(save_dir=model_path, file_name="test.html", show=False)
        assert os.stat(os.path.join(model_path, "test.png")).st_size > 1024 * 10
        assert os.stat(os.path.join(model_path, "test.html")).st_size > 1024 * 2


def test_cleanup():
    for model_file in txt_files:
        if model_file in skipped_files:
            continue
        model = model_file.split(".")[0]
        assert os.path.isdir(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                model,
            )
        )
        shutil.rmtree(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                model,
            )
        )



import os
import shutil

import matplotlib.pyplot as plt
import numpy as np

from biomass import create_model, run_simulation
from biomass.models import copy_to_current

MODEL_NAME: str = "insulin_signaling"

copy_to_current(MODEL_NAME)
assert os.path.exists(MODEL_NAME)
model = create_model(MODEL_NAME)


def test_simulate_successful():
    x = model.pval()
    y0 = model.ival()
    assert model.problem.simulate(x, y0) is None


def test_example_plot():
    assert run_simulation(model) is None
    res = np.load(os.path.join(model.path, "simulation_data", "simulations_original.npy"))

    plt.figure(figsize=(8, 8))
    plt.rcParams["font.family"] = "Arial"
    plt.rcParams["xtick.direction"] = "in"
    plt.rcParams["ytick.direction"] = "in"
    plt.rcParams["font.size"] = 18
    plt.rcParams["lines.linewidth"] = 4
    plt.rcParams["lines.markersize"] = 16

    plt.subplots_adjust(wspace=0.25, hspace=0.3)

    for i, obs_name in enumerate(model.observables):
        plt.subplot(2, 2, i + 1)
        for j, color in enumerate(
            ["darkblue", "cornflowerblue", "yellowgreen", "goldenrod", "brown"]
        ):
            plt.plot(model.problem.t, res[i, j], color=color)
        plt.xlim(0, 480)
        plt.xticks([0, 120, 240, 360, 480], fontweight="bold")
        plt.ylim(0, 1)
        plt.yticks([0, 0.2, 0.4, 0.6, 0.8, 1.0], fontweight="bold")
        plt.title(f"{obs_name}", fontweight="bold")
        if i % 2 == 0:
            plt.ylabel("Intensity (A.U.)", fontweight="bold")
        if i > 1:
            plt.xlabel("                           time (min)", fontweight="bold")
    plt.show()


def test_cleanup():
    shutil.rmtree(MODEL_NAME)



import os
import sys
from dataclasses import dataclass
from typing import Callable, Dict, Final, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from ..model_object import ModelObject
from ..plotting import SensitivityOptions
from .util import SignalingMetric, dlnyi_dlnxj, remove_nan


@dataclass
class ParameterSensitivity(SignalingMetric):
    """Sensitivity for parameters"""

    model: ModelObject
    create_metrics: Optional[Dict[str, Callable[[np.ndarray], Union[int, float]]]]

    def __post_init__(self) -> None:
        self._plotting: SensitivityOptions = self.model.viz.get_sensitivity_options()
        self._coefficients: Callable[[str], str] = lambda metric: os.path.join(
            self.model.path,
            "sensitivity_coefficients",
            "parameter",
            f"{metric}.npy",
        )
        self._path_to_figs: Callable[[str], str] = lambda metric: os.path.join(
            self.model.path,
            "figure",
            "sensitivity",
            "parameter",
            f"{metric}",
        )
        if self.create_metrics is not None:
            for name, function in self.create_metrics.items():
                self.quantification[name] = function

    def _get_param_indices(self, excluded_params: List[str]) -> List[int]:
        param_indices = []
        x = self.model.pval()
        for i, val in enumerate(x):
            if self.model.parameters[i] not in excluded_params and val != 0.0:
                param_indices.append(i)
        if not param_indices:
            raise ValueError("No nonzero parameters.")

        return param_indices

    def _calc_sensitivity_coefficients(
        self,
        metric: str,
        param_indices: List[int],
        show_progress: bool,
    ) -> np.ndarray:
        """Calculating Sensitivity Coefficients
        Parameters
        ----------
        metric : str
            The signaling metric used for sensitivity analysis.
        param_indices : list of int
            List of parameter indices for sensitivity analysis.
        Returns
        -------
        sensitivity_coefficients : numpy array
        """

        rate: Final[float] = 1.01  # 1% change
        n_file = self.model.get_executable()

        signaling_metric = np.full(
            (
                len(n_file),
                len(param_indices) + 1,
                len(self.model.observables),
                len(self.model.problem.conditions),
            ),
            np.nan,
        )
        for i, nth_paramset in enumerate(n_file):
            optimized = self.model.load_param(nth_paramset)
            for j, idx in enumerate(param_indices):
                x = optimized.params[:]
                x[idx] = optimized.params[idx] * rate
                if self.model.problem.simulate(x, optimized.initials) is None:
                    for k, _ in enumerate(self.model.observables):
                        for l, _ in enumerate(self.model.problem.conditions):
                            signaling_metric[i, j, k, l] = self.quantification[metric](
                                self.model.problem.simulations[k, l]
                            )
                if show_progress:
                    sys.stdout.write(
                        "\r{:d} / {:d}".format(
                            i * len(param_indices) + j + 1, len(n_file) * len(param_indices)
                        )
                    )
            # Signaling metric without perturbation (j=-1)
            x = optimized.params[:]
            if self.model.problem.simulate(x, optimized.initials) is None:
                for k, _ in enumerate(self.model.observables):
                    for l, _ in enumerate(self.model.problem.conditions):
                        signaling_metric[i, -1, k, l] = self.quantification[metric](
                            self.model.problem.simulations[k, l]
                        )
        sensitivity_coefficients = dlnyi_dlnxj(
            signaling_metric,
            len(n_file),
            len(param_indices),
            len(self.model.observables),
            len(self.model.problem.conditions),
            rate,
        )

        return sensitivity_coefficients

    def _load_sc(
        self,
        metric: str,
        param_indices: List[int],
        show_progress: bool,
    ) -> np.ndarray:
        """
        Load (or calculate) sensitivity coefficients.
        """
        if not os.path.isfile(self._coefficients(metric)):
            os.makedirs(
                os.path.join(
                    self.model.path,
                    "sensitivity_coefficients",
                    "parameter",
                ),
                exist_ok=True,
            )
            sensitivity_coefficients = self._calc_sensitivity_coefficients(
                metric, param_indices, show_progress
            )
            np.save(self._coefficients(metric), sensitivity_coefficients)
        else:
            sensitivity_coefficients = np.load(self._coefficients(metric))
            if len(param_indices) != sensitivity_coefficients.shape[1]:
                # User changed options['excluded_params'] after the last trial
                sensitivity_coefficients = self._calc_sensitivity_coefficients(
                    metric, param_indices, show_progress
                )
                np.save(self._coefficients(metric), sensitivity_coefficients)

        return sensitivity_coefficients

    def _barplot_sensitivity(
        self,
        metric: str,
        sensitivity_coefficients: np.ndarray,
        param_indices: List[int],
    ) -> None:
        """
        Visualize sensitivity coefficients using barplot.
        """
        os.makedirs(os.path.join(self._path_to_figs(metric), "barplot"), exist_ok=True)
        # rcParams
        self.model.viz.set_sensitivity_rcParams()

        for k, obs_name in enumerate(self.model.observables):
            plt.figure(figsize=self._plotting.figsize)
            plt.hlines([0], -self._plotting.width, len(param_indices), "k", lw=1)
            for l, condition in enumerate(self.model.problem.conditions):
                sensitivity_matrix = sensitivity_coefficients[:, :, k, l]
                nan_idx = []
                for m in range(sensitivity_matrix.shape[0]):
                    if np.isnan(sensitivity_matrix[m, :]).any():
                        nan_idx.append(m)
                sensitivity_matrix = np.delete(sensitivity_matrix, nan_idx, axis=0)
                if sensitivity_matrix.size != 0:
                    average = np.mean(sensitivity_matrix, axis=0)
                    if sensitivity_matrix.shape[0] == 1:
                        stdev = np.zeros(sensitivity_matrix.shape[1])
                    else:
                        stdev = np.std(sensitivity_matrix, axis=0, ddof=1)
                    plt.bar(
                        np.arange(len(param_indices)) + l * self._plotting.width,
                        average,
                        yerr=stdev,
                        ecolor=self._plotting.cmap[l],
                        capsize=2,
                        width=self._plotting.width,
                        color=self._plotting.cmap[l],
                        align="center",
                        label=condition,
                    )
            plt.xticks(
                np.arange(len(param_indices))
                + self._plotting.width * 0.5 * (len(self.model.problem.conditions) - 1),
                [self.model.parameters[i] for i in param_indices],
                fontsize=6,
                rotation=90,
            )
            plt.ylabel(
                "Control coefficients on\n" + metric + " (" + obs_name.replace("_", " ") + ")"
            )
            plt.xlim(-self._plotting.width, len(param_indices))
            if self._plotting.legend_kws is not None:
                plt.legend(**self._plotting.legend_kws)
            plt.savefig(
                os.path.join(
                    self._path_to_figs(metric),
                    "barplot",
                    f"{obs_name}",
                ),
            )
            plt.close()

    def _heatmap_sensitivity(
        self,
        metric: str,
        sensitivity_coefficients: np.ndarray,
        param_indices: List[int],
        clustermap_kws: dict,
        cbar_ax_tick_params: dict,
    ) -> None:
        """
        Visualize sensitivity coefficients using heatmap.
        """
        os.makedirs(os.path.join(self._path_to_figs(metric), "heatmap"), exist_ok=True)
        # rcParams
        self.model.viz.set_sensitivity_rcParams()

        for k, obs_name in enumerate(self.model.observables):
            for l, condition in enumerate(self.model.problem.conditions):
                sensitivity_matrix = remove_nan(sensitivity_coefficients[:, :, k, l])
                if sensitivity_matrix.shape[0] > 1 and not np.all(sensitivity_matrix == 0.0):
                    g = sns.clustermap(
                        data=sensitivity_matrix,
                        xticklabels=[self.model.parameters[i] for i in param_indices],
                        **clustermap_kws,
                    )
                    cbar = g.ax_heatmap.collections[0].colorbar
                    cbar.ax.tick_params(**cbar_ax_tick_params)
                    plt.setp(g.ax_heatmap.get_xticklabels(), rotation=90)
                    plt.savefig(
                        os.path.join(
                            self._path_to_figs(metric),
                            "heatmap",
                            f"{condition.replace(',', '')}_{obs_name}",
                        ),
                    )
                    plt.close()

    def analyze(
        self,
        *,
        metric: str,
        style: str,
        show_progress: bool,
        clustermap_kws: dict,
        cbar_ax_tick_params: dict,
        options: dict,
    ) -> None:
        """
        Perform sensitivity analysis.
        """
        if options["overwrite"] and os.path.isfile(self._coefficients(metric)):
            os.remove(self._coefficients(metric))
        param_indices = self._get_param_indices(options["excluded_params"])
        sensitivity_coefficients = self._load_sc(metric, param_indices, show_progress)
        if style == "barplot":
            self._barplot_sensitivity(
                metric,
                sensitivity_coefficients,
                param_indices,
            )
        elif style == "heatmap":
            if len(param_indices) < 2:
                raise ValueError(
                    "The number of parameters to be perturbed must be greater than 1, "
                    f"got {len(param_indices)}."
                )
            self._heatmap_sensitivity(
                metric,
                sensitivity_coefficients,
                param_indices,
                clustermap_kws,
                cbar_ax_tick_params,
            )
        else:
            raise ValueError("Available styles are: 'barplot', 'heatmap'")



import os
import shutil

import matplotlib.pyplot as plt
import numpy as np

from biomass import create_model, run_simulation
from biomass.models import copy_to_current

MODEL_NAME: str = "circadian_clock"

copy_to_current(MODEL_NAME)
assert os.path.exists(MODEL_NAME)
model = create_model(MODEL_NAME)


def test_simulate_successful():
    x = model.pval()
    y0 = model.ival()
    assert model.problem.simulate(x, y0) is None


def test_example_plot():
    assert run_simulation(model, viz_type="original") is None
    res = np.load(os.path.join(model.path, "simulation_data", "simulations_original.npy"))

    plt.rcParams["font.size"] = 16
    fig, ax1 = plt.subplots(figsize=(6, 4))
    ax2 = ax1.twinx()

    ax1.plot(model.problem.t, res[model.observables.index("Cry_mRNA"), 0], "c")
    ax1.plot(model.problem.t, res[model.observables.index("Per_mRNA"), 0], "m")
    ax1.set_xlim([0, 72])
    ax1.set_xticks([0, 12, 24, 36, 48, 60, 72])
    ax1.set_xlabel("Time (h)")
    ax1.set_ylim([0, 5])
    ax1.set_ylabel(
        r"$\it{Per}$"
        + " "
        + r"$\sf{(M_P)}$"
        + " and "
        + r"$\it{Cry}$"
        + " "
        + r"$\sf{(M_C)}$"
        + "\nmRNAs, nM"
    )
    ax2.plot(model.problem.t, res[model.observables.index("Bmal1_mRNA"), 0], "y")
    ax2.set_ylim([7, 10])
    ax2.set_yticks([7, 8, 9, 10])
    ax2.set_ylabel(r"$\it{Bmal1}$" + " mRNA " + r"$\sf{(M_B)}$" + ", nM")
    plt.show()


def test_cleanup():
    shutil.rmtree(MODEL_NAME)



import re
import sys
from collections import defaultdict
from dataclasses import dataclass, field
from difflib import SequenceMatcher
from typing import Dict, List, NamedTuple, Optional

import numpy as np

from .thermodynamic_restrictions import ComplexFormation, DuplicateError, ThermodynamicRestrictions


class KineticInfo(NamedTuple):
    reactants: tuple
    products: tuple
    modifiers: tuple
    rate: str


class DetectionError(Exception):
    pass


class ArrowError(ValueError):
    pass


PREPOSITIONS: List[str] = [
    "to",
    "for",
    "from",
    "up",
    "down",
    "in",
    "on",
    "at",
    "off",
    "into",
    "around",
    "among",
    "between",
    "of",
    "over",
    "above",
    "below",
    "under",
    "through",
    "across",
    "along",
    "near",
    "by",
    "beside",
]


@dataclass
class ReactionRules(ThermodynamicRestrictions):
    """Create an executable biochemical model from text.

    .. list-table:: Available reaction rules
        :widths: 25 50 25
        :header-rows: 1

        * - Rule
          - Example sentence
          - Parameters (optional)
        * - :func:`~pasmopy.construction.reaction_rules.dimerize`
          - *A* dimerizes <--> *AA*
          - .. math:: kf, kr
        * - :func:`~pasmopy.construction.reaction_rules.bind`
          - *A* binds *B* <--> *AB*
          - .. math:: kf, kr
        * - :func:`~pasmopy.construction.reaction_rules.dissociate`
          - *AB* dissociates to *A* and *B*
          - .. math:: kf, kr
        * - :func:`~pasmopy.construction.reaction_rules.is_phosphorylated`
          - *uA* is phosphorylated <--> *pA*
          - .. math:: kf, kr
        * - :func:`~pasmopy.construction.reaction_rules.is_dephosphorylated`
          - *pA* is dephosphorylated --> *uA*
          - .. math:: V, K
        * - :func:`~pasmopy.construction.reaction_rules.phosphorylate`
          - *B* phosphorylates *uA* --> *pA*
          - .. math:: V, K
        * - :func:`~pasmopy.construction.reaction_rules.dephosphorylate`
          - *B* dephosphorylates *pA* --> *uA*
          - .. math:: V, K
        * - :func:`~pasmopy.construction.reaction_rules.transcribe`
          - *B* transcribes *a*
          - .. math:: V, K, n, (KF, nF)
        * - :func:`~pasmopy.construction.reaction_rules.synthesize`
          - *B* synthesizes *A*
          - .. math:: kf
        * - :func:`~pasmopy.construction.reaction_rules.is_synthesized`
          - *A* is synthesized
          - .. math:: kf
        * - :func:`~pasmopy.construction.reaction_rules.degrade`
          - *B* degrades *A*
          - .. math:: kf
        * - :func:`~pasmopy.construction.reaction_rules.is_degraded`
          - *A* is degraded
          - .. math:: kf
        * - :func:`~pasmopy.construction.reaction_rules.translocate`
          - *Acyt* translocates from cytoplasm to nucleus (Vcyt, Vnuc) <--> *Anuc*
          - .. math:: kf, kr, (V_{pre}, V_{post})
        * - :func:`~pasmopy.construction.reaction_rules.user_defined`
          - @rxn Reactant --> Product: *define rate equation here*
          - *user-defined*

    From v0.2.2, you can specify directionality in binding-dissociation reaction via different arrows:

    .. code-block:: python

        E + S ⇄ ES | kf=0.003, kr=0.001 | E=100, S=50  # bi-directional
        ES → E + P | kf=0.002  # unidirectional

    Attributes
    ----------
    input_txt : str
        Model description file (*.txt), e.g.,
        `Kholodenko1999.txt <https://github.com/pasmopy/pasmopy/blob/master/tests/text_files/Kholodenko1999.txt>`_.
    similarity_threshold : float
        If all match_scores are below this value, expected_word will not be returned.
    parameters : list of strings
        ``x`` : model parameters.
    species : list of strings
        ``y`` : model species.
    reactions : list of strings
        ``v`` : flux vector.
    differential_equations : list of strings
        ``dydt`` : right-hand side of the differential equation.
    obs_desc : list of List[str]
        Description of observables.
    param_info : list of strings
        Information about parameter values.
    init_info : list of strings
        Information about initial values.
    param_constraints : list of strings
        Information about parameter constraints.
    param_excluded : list of strings
        List of parameters excluded from search params because of parameter constraints.
    fixed_species : list of strings
        List of species which should be held fixed (never consumed) during simulation.
    sim_tspan : list of strings ['t0', 'tf']
        Interval of integration.
    sim_conditions : list of List[str]
        Simulation conditions with stimulation.
    sim_unperturbed : str
        Untreated conditions to get steady state.
    rule_words : dict
        Words to identify reaction rules.
    nothing : List[str]
        Available symbol for degradation/creation to/from nothing.
    fwd_arrows : List[str]
        Available arrows for unidirectional reactions.
    double_arrows : List[str]
        Available arrows for bi-directional reactions.

    """

    input_txt: str
    similarity_threshold: float

    parameters: List[str] = field(
        default_factory=list,
        init=False,
    )
    species: List[str] = field(
        default_factory=list,
        init=False,
    )
    reactions: List[str] = field(
        default_factory=list,
        init=False,
    )
    kinetics: List[KineticInfo] = field(
        default_factory=list,
        init=False,
    )
    differential_equations: List[str] = field(
        default_factory=list,
        init=False,
    )
    obs_desc: List[List[str]] = field(
        default_factory=list,
        init=False,
    )
    param_info: List[str] = field(
        default_factory=list,
        init=False,
    )
    init_info: List[str] = field(
        default_factory=list,
        init=False,
    )
    param_constraints: List[str] = field(
        default_factory=list,
        init=False,
    )
    param_excluded: List[str] = field(
        default_factory=list,
        init=False,
    )
    fixed_species: List[str] = field(
        default_factory=list,
        init=False,
    )
    # Information about simulation
    sim_tspan: List[str] = field(
        default_factory=list,
        init=False,
    )
    sim_conditions: List[List[str]] = field(
        default_factory=list,
        init=False,
    )
    sim_unperturbed: str = field(
        default_factory=str,
        init=False,
    )
    # Words to identify reaction rules
    rule_words: Dict[str, List[str]] = field(
        default_factory=lambda: dict(
            _bind_and_dissociate=[
                " +",
            ],
            dimerize=[
                " dimerizes",
                " homodimerizes",
                " forms a dimer",
                " forms dimers",
            ],
            bind=[
                " binds",
                " forms complexes with",
            ],
            dissociate=[
                " is dissociated into",
                " dissociates to",
            ],
            is_phosphorylated=[
                " is phosphorylated",
            ],
            is_dephosphorylated=[
                " is dephosphorylated",
            ],
            phosphorylate=[
                " phosphorylates",
            ],
            dephosphorylate=[
                " dephosphorylates",
            ],
            transcribe=[
                " transcribe",
                " transcribes",
            ],
            synthesize=[
                " synthesizes",
                " promotes synthesis of",
                " increases",
                " is translated into",
            ],
            is_synthesized=[
                " is synthesized",
            ],
            degrade=[
                " degrades",
                " promotes degradation of",
                " decreases",
            ],
            is_degraded=[
                " is degraded",
            ],
            translocate=[
                " translocates",
                " is translocated",
            ],
        ),
        init=False,
    )
    nothing: List[str] = field(
        default_factory=lambda: ["∅", "0"],
        init=False,
    )
    fwd_arrows: List[str] = field(
        default_factory=lambda: [
            " → ",
            " ↣ ",
            " ↦ ",
            " ⇾ ",
            " ⟶ ",
            " ⟼ ",
            " ⥟ ",
            " ⥟ ",
            " ⇀ ",
            " ⇁ ",
            " ⇒ ",
            " ⟾ ",
        ]
        + [" {}> ".format("-" * i) for i in range(3)],
        init=False,
    )
    double_arrows: List[str] = field(
        default_factory=lambda: [
            " ↔ ",
            " ⟷ ",
            " ⇄ ",
            " ⇆ ",
            " ⇌ ",
            " ⇋ ",
            " ⇔ ",
            " ⟺ ",
        ]
        + [" <{}> ".format("-" * i) for i in range(3)],
        init=False,
    )

    def __post_init__(self) -> None:
        if not 0.0 < self.similarity_threshold < 1.0:
            raise ValueError("similarity_threshold must lie within (0, 1).")

    @staticmethod
    def _isfloat(string: str) -> bool:
        """
        Checking if a string can be converted to float.
        """
        try:
            float(string)
            return True
        except ValueError:
            return False

    @staticmethod
    def _remove_prefix(text: str, prefix: str) -> str:
        """
        Remove prefix from a text.
        """
        if text.startswith(prefix):
            return text[len(prefix) :]
        assert False

    def _available_arrows(self) -> List[str]:
        """
        Return all available arrow types.
        """
        return self.fwd_arrows + self.double_arrows

    def _set_params(self, line_num: Optional[int], func_name: Optional[str], *args: str) -> None:
        """
        Set model parameters.
        """
        for p_name in args:
            p_name_used = p_name + (
                f"{line_num:d}"
                if not p_name[-1].isdecimal()
                and isinstance(line_num, int)
                and func_name != "user_defined"
                else ""
            )
            if p_name_used not in self.parameters:
                self.parameters.append(p_name_used)

    def _set_species(self, *args: str) -> None:
        """
        Set model species.
        """
        for s_name in args:
            if s_name not in self.species:
                self.species.append(s_name)

    def _raise_exception(self, line_num: int, line: str) -> None:
        """
        Apply `state_transition` rule or raise `DetectionError` when a keyword is invalid.
        """
        expected_word = self._get_partial_similarity(line)
        if expected_word is None and any(
            map(lambda arrow: arrow in line, self._available_arrows())
        ):
            self.state_transition(line_num, line)
        else:
            raise DetectionError(
                f"Unregistered words in line{line_num:d}: {line}"
                + (f"\nMaybe: '{expected_word.lstrip()}'." if expected_word is not None else "")
            )

    def _process_pval_section(self, func_name: str, line_num: int, line: str, *args: str) -> None:
        param_values = line.split("|")[1].strip().split(",")
        if all("=" in pval for pval in param_values):
            for pval in param_values:
                base_param = pval.split("=")[0].strip(" ")
                if base_param.startswith("const "):
                    # Parameter names with 'const' will be added to param_excluded.
                    base_param = base_param.split("const ")[-1]
                    fixed = True
                else:
                    fixed = False
                if base_param in args:
                    if self._isfloat(pval.split("=")[1].strip(" ")):
                        self.param_info.append(
                            "x[C."
                            + base_param
                            + (f"{line_num:d}]" if func_name != "user_defined" else "]")
                            + " = "
                            + pval.split("=")[1].strip(" ")
                        )
                        # If a parameter value is initialized to 0.0 or fixed,
                        # then add it to param_excluded.
                        if float(pval.split("=")[1].strip(" ")) == 0.0 or fixed:
                            self.param_excluded.append(
                                base_param
                                + (f"{line_num:d}" if func_name != "user_defined" else "")
                            )
                    else:
                        raise ValueError(
                            f"line{line_num:d}: Parameter value must be int or float."
                        )
                else:
                    raise ValueError(
                        f"line{line_num:d}: '{pval.split('=')[0].strip(' ')}'\n"
                        f"Available parameters are: {', '.join(args)}."
                    )
        elif param_values[0].strip(" ").isdecimal():
            # Parameter constraints
            for param_name in args:
                # base_pname = self._get_base_pname(param_name)
                if (
                    f"{self._get_base_pname(param_name)}{int(param_values[0]):d}"
                ) not in self.parameters:
                    raise ValueError(
                        f"Line {line_num:d} and {int(param_values[0]):d} : "
                        "Different reaction rules in parameter constraints."
                    )
                else:
                    if f"x[C.{param_name}" + (
                        f"{line_num:d}]" if func_name != "user_defined" else "]"
                    ) != (
                        f"x[C.{self._get_base_pname(param_name)}" + f"{int(param_values[0]):d}]"
                    ):
                        self.param_excluded.append(
                            f"{param_name}"
                            + (f"{line_num:d}" if func_name != "user_defined" else "")
                        )
                        self.param_info.append(
                            f"x[C.{param_name}"
                            + (f"{line_num:d}]" if func_name != "user_defined" else "]")
                            + " = "
                            + f"x[C.{self._get_base_pname(param_name)}"
                            + f"{int(param_values[0]):d}]"
                        )
                        self.param_constraints.append(
                            f"x[C.{param_name}"
                            + (f"{line_num:d}]" if func_name != "user_defined" else "]")
                            + " = "
                            + f"x[C.{self._get_base_pname(param_name)}"
                            + f"{int(param_values[0]):d}]"
                        )
        else:
            raise ValueError(
                f"line{line_num:d}: {line}\nInvalid expression in the input parameter."
            )

    def _process_ival_section(self, line_num: int, line: str) -> None:
        initial_values = line.split("|")[2].strip().split(",")
        for ival in initial_values:
            if ival.startswith("fixed "):
                ival = ival.split("fixed ")[-1]
                self.fixed_species.append(ival.split("=")[0].strip(" "))
            if ival.split("=")[0].strip(" ") in line.split("|")[0]:
                if self._isfloat(ival.split("=")[1].strip(" ")):
                    self.init_info.append(
                        "y0[V."
                        + ival.split("=")[0].strip(" ")
                        + "] = "
                        + ival.split("=")[1].strip(" ")
                    )
                else:
                    raise ValueError(f"line{line_num:d}: Initial value must be int or float.")
            else:
                raise NameError(
                    f"line{line_num:d}: " f"Name'{ival.split('=')[0].strip(' ')}' is not defined."
                )

    @staticmethod
    def _get_base_pname(param_name: str) -> str:
        while param_name[-1].isdecimal():
            param_name = param_name[:-1]
        return param_name

    def _preprocessing(self, func_name: str, line_num: int, line: str, *args: str) -> List[str]:
        """
        Extract the information about parameter and/or initial values
        if '|' in the line and find a keyword to identify reaction rules.

        Parameters
        ----------
        func_name : str
            Name of the rule function.

        line_num : int
            Line number.

        line : str
            Each line of the input text.

        Returns
        -------
        description : list of strings

        """
        self._set_params(line_num, func_name, *args)
        if "|" in line:
            if line.split("|")[1].strip():
                self._process_pval_section(func_name, line_num, line, *args)
            if line.count("|") > 1 and line.split("|")[2].strip():
                self._process_ival_section(line_num, line)
            line = line.split("|")[0]
        if func_name not in ["state_transition", "user_defined"]:
            hit_words: List[str] = []
            for word in self.rule_words[func_name]:
                # Choose longer word
                if word in line:
                    hit_words.append(word)
            description = line.strip().split(max(hit_words, key=len))
            if description[1] and not description[1].startswith(" "):
                self._raise_exception(line_num, line)
        elif func_name == "state_transition":
            description = [line.strip()]
        elif func_name == "user_defined":
            description = line.strip().split(":")
        else:
            assert False
        return description

    @staticmethod
    def _word2scores(word: str, sentence: str) -> List[float]:
        """
        Calculate similarity scores between word and sentence.

        Parameters
        ----------
        word : str
            User-defined word.
        sentence : str
            Textual unit consisting of two or more words.

        returns
        -------
        scores : list
            List containing similarity scores.

        """
        scores = [
            SequenceMatcher(None, word, sentence[i : i + len(word)]).ratio()
            for i in range(len(sentence) - len(word) + 1)
        ]
        return scores

    def _get_partial_similarity(self, line: str) -> Optional[str]:
        """
        Suggest similar rule word when user-defined word is not registered
        in rule_words.

        Parameters
        ----------
        line : str
            Each line of the input text.

        Returns
        -------
        expected_word : Optional[str]
            Rule word with the highest similarity score.

        """
        match_words = []
        match_scores = []
        str_subset = []
        for rules in self.rule_words.values():
            for word in rules:
                ratio = self._word2scores(word, line)
                if ratio:
                    match_words.append(word)
                    match_scores.append(max(ratio))
                    str_subset.append(line[np.argmax(ratio) : np.argmax(ratio) + len(word)])
        expected_word = (
            None
            if all([score < self.similarity_threshold for score in match_scores])
            else match_words[np.argmax(match_scores)]
        )
        # original_word = (
        #     None if expected_word is None else str_subset[match_words.index(expected_word)]
        # )
        return expected_word

    @staticmethod
    def _remove_prepositions(sentence: str) -> str:
        """
        Remove preposition from text not to use it for identifying reaction rules.
        """
        for preposition in PREPOSITIONS:
            if sentence.endswith(f" {preposition}"):
                return sentence[: -len(preposition) - 1]
        return sentence

    def _get_arrow_error_message(self, line_num: int) -> str:
        message = (
            f"line{line_num}: Use one of ({', '.join(self.fwd_arrows)}) for unidirectional "
            f"reaction or ({', '.join(self.double_arrows)}) for bi-directional reaction"
        )
        return message

    def _redirect_rules(
        self,
        line_num: int,
        is_binding: bool,
        is_unidirectional: bool,
        c1: str,
        c2: str,
        cmplx: str,
    ) -> None:
        if c1 == c2:
            # A + A --> AA
            self.dimerize(
                line_num, f"{c1} dimerizes {'-->' if is_unidirectional else '<-->'} {cmplx}"
            )
        elif c1 == cmplx and c2 != cmplx:
            if is_binding:
                # A + B --> A
                self.degrade(line_num, f"{c1} degrades {c2}")
            else:
                # A --> A + B
                self.synthesize(line_num, f"{c1} synthesizes {c2}")
        elif c1 != cmplx and c2 == cmplx:
            if is_binding:
                # A + B --> B
                self.degrade(line_num, f"{c2} degrades {c1}")
            else:
                # B --> A + B
                self.synthesize(line_num, f"{c2} synthesizes {c1}")

    def _bind_and_dissociate(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'A + B --> AB'  # bind, unidirectional
        >>> 'AB --> A + B'  # dissociate, unidirectional
        >>> 'A + B <--> AB' # bind and dissociate, bidirectional
        """
        for arrow in self._available_arrows():
            if arrow in line:
                params_used = ["kf"] if arrow in self.fwd_arrows else ["kf", "kr"]
                break
        else:
            raise ArrowError(self._get_arrow_error_message(line_num) + ".")
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, *params_used
        )
        if len(description) != 2:
            # In the case like: E + S --> E + P
            self.state_transition(line_num, line)
            return
        is_binding: bool
        is_unidirectional: bool
        for arrow in self._available_arrows():
            if arrow in description[1]:
                is_binding = True
                is_unidirectional = True if arrow in self.fwd_arrows else False
                component1 = description[0].strip(" ")
                component2 = description[1].split(arrow)[0].strip(" ")
                complex = description[1].split(arrow)[1].strip(" ")
                break
            elif arrow in description[0]:
                is_binding = False
                is_unidirectional = True if arrow in self.fwd_arrows else False
                component1 = description[0].split(arrow)[1].strip(" ")
                component2 = description[1].strip(arrow)
                complex = description[0].split(arrow)[0].strip(" ")
                break
        else:
            raise ArrowError(self._get_arrow_error_message(line_num) + ".")
        if (component1 == complex or component2 == complex) and not is_unidirectional:
            raise ValueError(f"line{line_num:d}: {complex} <- Use a different name.")
        elif (
            (component1 == component2)
            or (component1 != complex and component2 == complex and is_unidirectional)
            or (component1 == complex and component2 != complex and is_unidirectional)
        ):
            self._redirect_rules(
                line_num, is_binding, is_unidirectional, component1, component2, complex
            )
            return
        else:
            self._set_species(component1, component2, complex)
            self.complex_formations.append(
                ComplexFormation(line_num, set([component1, component2]), complex, is_binding)
            )
            if is_binding:
                self.reactions.append(
                    f"v[{line_num:d}] = "
                    f"x[C.kf{line_num:d}] * y[V.{component1}] * y[V.{component2}]"
                    + (f" - x[C.kr{line_num:d}] * y[V.{complex}]" if not is_unidirectional else "")
                )
                self.kinetics.append(
                    KineticInfo(
                        (component1, component2),
                        (complex,),
                        (),
                        f"kf{line_num:d} * {component1} * {component2}",
                    )
                )
                if not is_unidirectional:
                    self.kinetics.append(
                        KineticInfo(
                            (complex,), (component1, component2), (), f"kr{line_num:d} * {complex}"
                        )
                    )
            else:
                self.reactions.append(
                    f"v[{line_num:d}] = "
                    f"x[C.kf{line_num:d}] * y[V.{complex}]"
                    + (
                        f" - x[C.kr{line_num:d}] * y[V.{component1}] * y[V.{component2}]"
                        if not is_unidirectional
                        else ""
                    )
                )
                self.kinetics.append(
                    KineticInfo(
                        (complex,), (component1, component2), (), f"kf{line_num:d} * {complex}"
                    )
                )
                if not is_unidirectional:
                    self.kinetics.append(
                        KineticInfo(
                            (component1, component2),
                            (complex,),
                            (),
                            f"kr{line_num:d} * {component1} * {component2}",
                        )
                    )

            counter_component1, counter_component2, counter_complex = (0, 0, 0)
            for i, eq in enumerate(self.differential_equations):
                if f"dydt[V.{component1}]" in eq:
                    counter_component1 += 1
                    self.differential_equations[i] = (
                        eq + (" - " if is_binding else " + ") + f"v[{line_num:d}]"
                    )
                elif f"dydt[V.{component2}]" in eq:
                    counter_component2 += 1
                    self.differential_equations[i] = (
                        eq + (" - " if is_binding else " + ") + f"v[{line_num:d}]"
                    )
                elif f"dydt[V.{complex}]" in eq:
                    counter_complex += 1
                    self.differential_equations[i] = (
                        eq + (" + " if is_binding else " - ") + f"v[{line_num:d}]"
                    )
            if counter_component1 == 0:
                self.differential_equations.append(
                    f"dydt[V.{component1}] = - v[{line_num:d}]"
                    if is_binding
                    else f"dydt[V.{component1}] = + v[{line_num:d}]"
                )
            if counter_component2 == 0:
                self.differential_equations.append(
                    f"dydt[V.{component2}] = - v[{line_num:d}]"
                    if is_binding
                    else f"dydt[V.{component2}] = + v[{line_num:d}]"
                )
            if counter_complex == 0:
                self.differential_equations.append(
                    f"dydt[V.{complex}] = + v[{line_num:d}]"
                    if is_binding
                    else f"dydt[V.{complex}] = - v[{line_num:d}]"
                )

    def dimerize(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'A dimerizes <--> AA'
        >>> 'A homodimerizes <--> AA'
        >>> 'A forms a dimer <--> AA'
        >>> 'A forms dimers <--> AA'

        Notes
        -----
        * Parameters
            .. math:: kf, kr

        * Rate equation
            .. math:: v = kf * [A] * [A] - kr * [AA]

        * Differential equation
            .. math::

                d[A]]/dt = - 2 * v

                d[AA]/dt = + v

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "kf", "kr"
        )
        is_unidirectional: bool
        monomer = description[0].strip(" ")
        for arrow in self._available_arrows():
            if arrow in description[1]:
                is_unidirectional = True if arrow in self.fwd_arrows else False
                dimer = description[1].split(arrow)[1].strip(" ")
                break
        else:
            raise ArrowError(
                self._get_arrow_error_message(line_num) + " to specify the name of the dimer."
            )
        if monomer == dimer:
            raise ValueError(f"{dimer} <- Use a different name.")
        self._set_species(monomer, dimer)
        self.complex_formations.append(ComplexFormation(line_num, set(monomer), dimer, True))
        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.kf{line_num:d}] * y[V.{monomer}] * y[V.{monomer}]"
            + (f" - x[C.kr{line_num:d}] * y[V.{dimer}]" if not is_unidirectional else "")
        )
        self.kinetics.append(
            KineticInfo(
                (monomer, monomer),
                (dimer,),
                (),
                f"kf{line_num:d} * {monomer} * {monomer}",
            )
        )
        if not is_unidirectional:
            self.kinetics.append(
                KineticInfo(
                    (dimer,),
                    (monomer, monomer),
                    (),
                    f"kr{line_num:d} * {dimer}",
                )
            )
        counter_monomer, counter_dimer = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{monomer}]" in eq:
                counter_monomer += 1
                self.differential_equations[i] = eq + f" - 2 * v[{line_num:d}]"
            elif f"dydt[V.{dimer}]" in eq:
                counter_dimer += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_monomer == 0:
            self.differential_equations.append(f"dydt[V.{monomer}] = - v[{line_num:d}]")
        if counter_dimer == 0:
            self.differential_equations.append(f"dydt[V.{dimer}] = + v[{line_num:d}]")

    def bind(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'A binds B <--> AB'
        >>> 'A forms complexes with B <--> AB'

        Notes
        -----
        * Parameters
            .. math:: kf, kr

        * Rate equation
            .. math:: v = kf * [A] * [B] - kr * [AB]

        * Differential equation
            .. math::

                d[A]/dt = - v

                d[B]/dt = - v

                d[AB]/dt = + v

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "kf", "kr"
        )
        is_unidirectional: bool
        component1 = description[0].strip(" ")
        for arrow in self._available_arrows():
            if arrow in description[1]:
                is_unidirectional = True if arrow in self.fwd_arrows else False
                component2 = description[1].split(arrow)[0].strip(" ")
                complex = description[1].split(arrow)[1].strip(" ")
                break
        else:
            raise ArrowError(
                self._get_arrow_error_message(line_num)
                + " to specify the name of the protein complex."
            )
        if component1 == complex or component2 == complex:
            raise ValueError(f"line{line_num:d}: {complex} <- Use a different name.")
        elif component1 == component2:
            raise ValueError(f"line{line_num}: {line}\nUse `dimerize()` rule instead of `bind()`.")
        else:
            self._set_species(component1, component2, complex)
            self.complex_formations.append(
                ComplexFormation(line_num, set([component1, component2]), complex, True)
            )
            self.reactions.append(
                f"v[{line_num:d}] = "
                f"x[C.kf{line_num:d}] * y[V.{component1}] * y[V.{component2}]"
                + (f" - x[C.kr{line_num:d}] * y[V.{complex}]" if not is_unidirectional else "")
            )
            self.kinetics.append(
                KineticInfo(
                    (component1, component2),
                    (complex,),
                    (),
                    f"kf{line_num:d} * {component1} * {component2}",
                )
            )
            if not is_unidirectional:
                self.kinetics.append(
                    KineticInfo(
                        (complex,),
                        (component1, component2),
                        (),
                        f"kr{line_num:d} * {complex}",
                    )
                )
            counter_component1, counter_component2, counter_complex = (0, 0, 0)
            for i, eq in enumerate(self.differential_equations):
                if f"dydt[V.{component1}]" in eq:
                    counter_component1 += 1
                    self.differential_equations[i] = eq + f" - v[{line_num:d}]"
                elif f"dydt[V.{component2}]" in eq:
                    counter_component2 += 1
                    self.differential_equations[i] = eq + f" - v[{line_num:d}]"
                elif f"dydt[V.{complex}]" in eq:
                    counter_complex += 1
                    self.differential_equations[i] = eq + f" + v[{line_num:d}]"
            if counter_component1 == 0:
                self.differential_equations.append(f"dydt[V.{component1}] = - v[{line_num:d}]")
            if counter_component2 == 0:
                self.differential_equations.append(f"dydt[V.{component2}] = - v[{line_num:d}]")
            if counter_complex == 0:
                self.differential_equations.append(f"dydt[V.{complex}] = + v[{line_num:d}]")

    def dissociate(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'AB dissociates to A and B'
        >>> 'AB is dissociated into A and B'

        Notes
        -----
        * Parameters
            .. math:: kf, kr

        * Rate equation
            .. math:: v = kf * [AB] - kr * [A] * [B]

        * Differential equation
            .. math::

                d[A]/dt = + v

                d[B]/dt = + v

                d[AB]/dt = - v

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "kf", "kr"
        )
        complex = description[0].strip(" ")
        if " and " not in description[1]:
            raise ValueError(
                f"Use 'and' in line{line_num:d}:\ne.g., AB is dissociated into A and B"
            )
        else:
            component1 = description[1].split(" and ")[0].strip(" ")
            component2 = description[1].split(" and ")[1].strip(" ")
        self._set_species(complex, component1, component2)
        self.complex_formations.append(
            ComplexFormation(line_num, set([component1, component2]), complex, False)
        )
        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.kf{line_num:d}] * y[V.{complex}]"
            f" - x[C.kr{line_num:d}] * y[V.{component1}] * y[V.{component2}]"
        )
        self.kinetics.append(
            KineticInfo((complex,), (component1, component2), (), f"kf{line_num:d} * {complex}")
        )
        self.kinetics.append(
            KineticInfo(
                (component1, component2),
                (complex,),
                (),
                f"kr{line_num:d} * {component1} * {component2}",
            )
        )
        counter_complex, counter_component1, counter_component2 = (0, 0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{complex}]" in eq:
                counter_complex += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif f"dydt[V.{component1}]" in eq:
                counter_component1 += 1
                self.differential_equations[i] = (
                    eq + f" + v[{line_num:d}]"
                    if component1 != component2
                    else eq + f" + 2 * v[{line_num:d}]"
                )
            elif f"dydt[V.{component2}]" in eq:
                counter_component2 += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_complex == 0:
            self.differential_equations.append(f"dydt[V.{complex}] = - v[{line_num:d}]")
        if counter_component1 == 0:
            self.differential_equations.append(f"dydt[V.{component1}] = + v[{line_num:d}]")
        if counter_component2 == 0:
            self.differential_equations.append(f"dydt[V.{component2}] = + v[{line_num:d}]")

    def is_phosphorylated(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'uA is phosphorylated <--> pA'

        Notes
        -----
        * Parameters
            .. math:: kf, kr

        * Rate equation
            .. math:: v = kf * [uA] - kr * [pA]

        * Differential equation
            .. math::

                d[uA]/dt = - v

                d[pA]/dt = + v

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "kf", "kr"
        )
        is_unidirectional: bool
        unphosphorylated_form = description[0].strip(" ")
        for arrow in self._available_arrows():
            if arrow in description[1]:
                is_unidirectional = True if arrow in self.fwd_arrows else False
                phosphorylated_form = description[1].split(arrow)[1].strip(" ")
                break
        else:
            raise ArrowError(
                self._get_arrow_error_message(line_num)
                + " to specify the name of the phosphorylated protein."
            )
        self._set_species(unphosphorylated_form, phosphorylated_form)

        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.kf{line_num:d}] * y[V.{unphosphorylated_form}]"
            + (
                f" - x[C.kr{line_num:d}] * y[V.{phosphorylated_form}]"
                if not is_unidirectional
                else ""
            )
        )
        self.kinetics.append(
            KineticInfo(
                (unphosphorylated_form,),
                (phosphorylated_form,),
                (),
                f"kf{line_num:d} * {unphosphorylated_form}",
            )
        )
        if not is_unidirectional:
            self.kinetics.append(
                KineticInfo(
                    (phosphorylated_form,),
                    (unphosphorylated_form,),
                    (),
                    f"kr{line_num:d} * {phosphorylated_form}",
                )
            )
        counter_unphosphorylated_form, counter_phosphorylated_form = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{unphosphorylated_form}]" in eq:
                counter_unphosphorylated_form += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif f"dydt[V.{phosphorylated_form}]" in eq:
                counter_phosphorylated_form += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_unphosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{unphosphorylated_form}] = - v[{line_num:d}]"
            )
        if counter_phosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{phosphorylated_form}] = + v[{line_num:d}]"
            )

    def is_dephosphorylated(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'pA is dephosphorylated --> uA'

        Notes
        -----
        * Parameters
            .. math:: V, K

        * Rate equation
            .. math:: v = V * [pA] / (K + [pA])

        * Differential equation
            .. math::

                d[uA]/dt = + v

                d[pA]/dt = - v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "V", "K")
        phosphorylated_form = description[0].strip(" ")
        for arrow in self.fwd_arrows:
            if arrow in description[1]:
                unphosphorylated_form = description[1].split(arrow)[1].strip(" ")
                break
        else:
            raise ArrowError(
                f"line{line_num:d}: Use one of ({', '.join(self.fwd_arrows)}) "
                "to specify the name of the dephosphorylated protein."
            )
        self._set_species(phosphorylated_form, unphosphorylated_form)

        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.V{line_num:d}] * y[V.{phosphorylated_form}] / "
            f"(x[C.K{line_num:d}] + y[V.{phosphorylated_form}])"
        )
        self.kinetics.append(
            KineticInfo(
                (phosphorylated_form,),
                (unphosphorylated_form,),
                (),
                f"V{line_num:d} * {phosphorylated_form} / "
                f"(K{line_num:d} + {phosphorylated_form})",
            )
        )
        counter_unphosphorylated_form, counter_phosphorylated_form = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{unphosphorylated_form}]" in eq:
                counter_unphosphorylated_form += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
            elif f"dydt[V.{phosphorylated_form}]" in eq:
                counter_phosphorylated_form += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
        if counter_unphosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{unphosphorylated_form}] = + v[{line_num:d}]"
            )
        if counter_phosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{phosphorylated_form}] = - v[{line_num:d}]"
            )

    def phosphorylate(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'B phosphorylates uA --> pA'

        Notes
        -----
        * Parameters
            .. math:: V, K

        * Rate equation
            .. math:: v = V * [B] * [uA] / (K + [uA])

        * Differential equation
            .. math::

                d[uA]/dt = - v

                d[pA]/dt = + v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "V", "K")
        kinase = description[0].strip(" ")
        for arrow in self.fwd_arrows:
            if arrow in description[1]:
                unphosphorylated_form = description[1].split(arrow)[0].strip(" ")
                phosphorylated_form = description[1].split(arrow)[1].strip(" ")
                break
        else:
            raise ValueError(
                f"line{line_num:d}: "
                f"Use one of {', '.join(self.fwd_arrows)} to specify "
                "the name of the phosphorylated (or activated) protein."
            )
        if unphosphorylated_form == phosphorylated_form:
            raise ValueError(f"line{line_num:d}: {phosphorylated_form} <- Use a different name.")
        self._set_species(kinase, unphosphorylated_form, phosphorylated_form)

        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.V{line_num:d}] * y[V.{kinase}] * y[V.{unphosphorylated_form}] / "
            f"(x[C.K{line_num:d}] + y[V.{unphosphorylated_form}])"
        )
        self.kinetics.append(
            KineticInfo(
                (unphosphorylated_form,),
                (phosphorylated_form,),
                (kinase,),
                f"V{line_num:d} * {kinase} * {unphosphorylated_form} / "
                f"(K{line_num:d} + {unphosphorylated_form})",
            )
        )
        counter_unphosphorylated_form, counter_phosphorylated_form = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{unphosphorylated_form}]" in eq:
                counter_unphosphorylated_form += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif f"dydt[V.{phosphorylated_form}]" in eq:
                counter_phosphorylated_form += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_unphosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{unphosphorylated_form}] = - v[{line_num:d}]"
            )
        if counter_phosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{phosphorylated_form}] = + v[{line_num:d}]"
            )

    def dephosphorylate(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'B dephosphorylates pA --> uA'

        Notes
        -----
        * Parameters
            .. math:: V, K

        * Rate equation
            .. math:: v = V * [B] * [pA] / (K + [pA])

        * Differential equation
            .. math::

                d[uA]/dt = + v

                d[pA]/dt = - v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "V", "K")
        phosphatase = description[0].strip(" ")
        for arrow in self.fwd_arrows:
            if arrow in description[1]:
                phosphorylated_form = description[1].split(arrow)[0].strip(" ")
                unphosphorylated_form = description[1].split(arrow)[1].strip(" ")
                break
        else:
            raise ArrowError(
                f"line{line_num:d}: "
                f"Use one of {', '.join(self.fwd_arrows)} to specify "
                "the name of the dephosphorylated (or deactivated) protein."
            )
        if phosphorylated_form == unphosphorylated_form:
            raise ValueError(f"line{line_num:d}: {unphosphorylated_form} <- Use a different name.")
        self._set_species(phosphatase, phosphorylated_form, unphosphorylated_form)

        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.V{line_num:d}] * y[V.{phosphatase}] * y[V.{phosphorylated_form}] / "
            f"(x[C.K{line_num:d}] + y[V.{phosphorylated_form}])"
        )
        self.kinetics.append(
            KineticInfo(
                (phosphorylated_form,),
                (unphosphorylated_form,),
                (phosphatase,),
                f"V{line_num:d} * {phosphatase} * {phosphorylated_form} / "
                f"(K{line_num:d} + {phosphorylated_form})",
            )
        )
        counter_phosphorylated_form, counter_unphosphorylated_form = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{phosphorylated_form}]" in eq:
                counter_phosphorylated_form += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif f"dydt[V.{unphosphorylated_form}]" in eq:
                counter_unphosphorylated_form += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_phosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{phosphorylated_form}] = - v[{line_num:d}]"
            )
        if counter_unphosphorylated_form == 0:
            self.differential_equations.append(
                f"dydt[V.{unphosphorylated_form}] = + v[{line_num:d}]"
            )

    def transcribe(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'B transcribes a'
        >>> 'B1 & B2 transcribe a'  # (AND-gate)
        >>> 'B transcribes a, repressed by C'  # (Negative regulation)

        Notes
        -----
        * Parameters
            .. math:: V, K, n, (KF, nF)

        * Rate equation
            .. math::

                v = V * [B] ^ {n} / (K ^ {n} + [B] ^ {n})

                v = V * ([B1] * [B2]) ^ {n} / (K ^ {n} + ([B1] * [B2]) ^ {n})

                v = V * [B] ^ {n} / (K ^ {n} + [B] ^ {n} + ([C] / KF) ^ {nF})

        * Differential equation
            .. math:: d[a]/dt = + v

        """
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, "V", "K", "n", "KF", "nF"
        )
        repressor: Optional[str] = None
        ratio = self._word2scores(", repressed by", description[1])
        if not ratio or max(ratio) < 1.0:
            self.parameters.remove(f"KF{line_num:d}")
            self.parameters.remove(f"nF{line_num:d}")
            mRNA = description[1].strip()
            if " " in mRNA:
                # Fix typo in line{line_num:d}
                raise ValueError(
                    f"line{line_num:d}: "
                    "Add ', repressed by XXX' to describe negative regulation from XXX."
                )
        else:
            # Add negative regulation from repressor
            mRNA = description[1].split(", repressed by")[0].strip()
            repressor = description[1].split(", repressed by")[1].strip()
        if " & " not in description[0]:
            TF = description[0].strip(" ")
            modifiers = (TF,)
            self._set_species(mRNA, TF)
            if repressor is not None:
                self._set_species(repressor)
                modifiers = modifiers + (repressor,)
            self.reactions.append(
                f"v[{line_num:d}] = "
                f"x[C.V{line_num:d}] * y[V.{TF}] ** x[C.n{line_num:d}] / "
                f"(x[C.K{line_num:d}] ** x[C.n{line_num:d}] + "
                f"y[V.{TF}] ** x[C.n{line_num:d}]"
                + (
                    ")"
                    if repressor is None
                    else f" + (y[V.{repressor}] / x[C.KF{line_num:d}]) ** x[C.nF{line_num:d}])"
                )
            )
            self.kinetics.append(
                KineticInfo(
                    (),
                    (mRNA,),
                    modifiers,
                    f"V{line_num:d} * {TF} ** n{line_num:d} / "
                    f"(K{line_num:d} ** n{line_num:d} + "
                    f"{TF} ** n{line_num:d}"
                    + (
                        ")"
                        if repressor is None
                        else f" + ({repressor} / KF{line_num:d}) ** nF{line_num:d})"
                    ),
                )
            )
        else:
            # AND-gate
            TFs = [TF.strip(" ") for TF in description[0].split(" & ")]
            self._set_species(mRNA, *TFs)
            modifiers = (*TFs,)
            if repressor is not None:
                self._set_species(repressor)
                modifiers = modifiers + (repressor,)
            self.reactions.append(
                f"v[{line_num:d}] = "
                f"x[C.V{line_num:d}] * ({'y[V.' + '] * y[V.'.join(TFs) + ']'}) ** x[C.n{line_num:d}] / "
                f"(x[C.K{line_num:d}] ** x[C.n{line_num:d}] + "
                f"({'y[V.' + '] * y[V.'.join(TFs) + ']'}) ** x[C.n{line_num:d}]"
                + (
                    ")"
                    if repressor is None
                    else f" + (y[V.{repressor}] / x[C.KF{line_num:d}]) ** x[C.nF{line_num:d}])"
                )
            )
            self.kinetics.append(
                KineticInfo(
                    (),
                    (mRNA,),
                    modifiers,
                    f"V{line_num:d} * ({' * '.join(TFs)}) ** n{line_num:d} / "
                    f"(K{line_num:d} ** n{line_num:d} + "
                    f"({' * '.join(TFs)}) ** n{line_num:d}"
                    + (
                        ")"
                        if repressor is None
                        else f" + ({repressor} / KF{line_num:d}) ** nF{line_num:d})"
                    ),
                )
            )
        counter_mRNA = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{mRNA}]" in eq:
                counter_mRNA += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_mRNA == 0:
            self.differential_equations.append(f"dydt[V.{mRNA}] = + v[{line_num:d}]")

    def synthesize(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'B synthesizes A'

        Notes
        -----
        * Parameters
            .. math:: kf

        * Rate equation
            .. math:: v = kf * [B]

        * Differential equation
            .. math:: d[A]/dt = + v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "kf")
        catalyst = description[0].strip(" ")
        product = description[1].strip(" ")
        self._set_species(catalyst, product)
        self.reactions.append(f"v[{line_num:d}] = x[C.kf{line_num:d}] * y[V.{catalyst}]")
        self.kinetics.append(
            KineticInfo((), (product,), (catalyst,), f"kf{line_num:d} * {catalyst}")
        )
        counter_product = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{product}]" in eq:
                counter_product += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_product == 0:
            self.differential_equations.append(f"dydt[V.{product}] = + v[{line_num:d}]")

    def is_synthesized(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'A is synthesized'

        Notes
        -----
        * Parameters
            .. math:: kf

        * Rate equation
            .. math:: v = kf

        * Differential equation
            .. math:: d[A]/dt = + v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "kf")
        chemical_species = description[0].strip(" ")
        self._set_species(chemical_species)
        self.reactions.append(f"v[{line_num:d}] = x[C.kf{line_num:d}]")
        self.kinetics.append(KineticInfo((), (chemical_species,), (), f"kf{line_num:d}"))
        counter_chemical_species = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{chemical_species}]" in eq:
                counter_chemical_species += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_chemical_species == 0:
            self.differential_equations.append(f"dydt[V.{chemical_species}] = + v[{line_num:d}]")

    def degrade(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'B degrades A'

        Notes
        -----
        * Parameters
            .. math:: kf

        * Rate equation
            .. math:: v = kf * [B] * [A]

        * Differential equation
            .. math:: d[A]/dt = - v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "kf")
        protease = description[0].strip(" ")
        protein = description[1].strip(" ")
        self._set_species(protease, protein)
        self.reactions.append(
            f"v[{line_num:d}] = x[C.kf{line_num:d}] * y[V.{protease}] * y[V.{protein}]"
        )
        self.kinetics.append(
            KineticInfo((protein,), (), (protease,), f"kf{line_num:d} * {protease} * {protein}")
        )
        counter_protein = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{protein}]" in eq:
                counter_protein += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
        if counter_protein == 0:
            self.differential_equations.append(f"dydt[V.{protein}] = - v[{line_num:d}]")

    def is_degraded(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> 'A is degraded'

        Notes
        -----
        * Parameters
            .. math:: kf

        * Rate equation
            .. math:: v = kf * [A]

        * Differential equation
            .. math:: d[A]/dt = - v

        """
        description = self._preprocessing(sys._getframe().f_code.co_name, line_num, line, "kf")
        chemical_species = description[0].strip(" ")
        self._set_species(chemical_species)
        self.reactions.append(f"v[{line_num:d}] = x[C.kf{line_num:d}] * y[V.{chemical_species}]")
        self.kinetics.append(
            KineticInfo((chemical_species,), (), (), f"kf{line_num:d} * {chemical_species}")
        )
        counter_chemical_species = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{chemical_species}]" in eq:
                counter_chemical_species += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
        if counter_chemical_species == 0:
            self.differential_equations.append(f"dydt[V.{chemical_species}] = - v[{line_num:d}]")

    def translocate(self, line_num: int, line: str) -> None:
        r"""
        Examples
        --------
        >>> 'A_at_cyt translocates from cytoplasm to nucleus (V_cyt, V_nuc) <--> A_at_nuc'
        >>> 'A_at_cyt is translocated from cytoplasm to nucleus (V_cyt, V_nuc) <--> A_at_nuc'

        Notes
        -----
        * Parameters
            .. math:: kf, kr, (V_{pre}, V_{post})

        * Rate equation
            .. math:: v = kf * [A\_at\_pre] - kr * (V_{post} / V_{pre}) * [A\_at\_post]

        * Differential equation
            .. math::

                d[A\_at\_pre]/dt = - v

                d[A\_at\_post]/dt = + v * (V_{pre} / V_{post})

        """
        for arrow in self._available_arrows():
            if arrow in line:
                is_unidirectional = True if arrow in self.fwd_arrows else False
                params_used = ["kf"] if is_unidirectional else ["kf", "kr"]
                break
        else:
            raise ArrowError(self._get_arrow_error_message(line_num))
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, *params_used
        )
        pre_translocation = description[0].strip(" ")
        for arrow in self._available_arrows():
            if arrow in description[1]:
                post_translocation = description[1].split(arrow)[1].strip(" ")
                break
        else:
            assert False
        if pre_translocation == post_translocation:
            raise ValueError(f"line{line_num:d}: {post_translocation} <- Use a different name.")
        # Information about compartment volumes
        if "(" in description[1] and ")" in description[1]:
            [pre_volume, post_volume] = description[1].split("(")[-1].split(")")[0].split(",")
            if not self._isfloat(pre_volume.strip(" ")) or not self._isfloat(
                post_volume.strip(" ")
            ):
                raise ValueError("pre_volume and post_volume must be float or int.")
        else:
            [pre_volume, post_volume] = ["1", "1"]
        self._set_species(pre_translocation, post_translocation)
        self.reactions.append(
            f"v[{line_num:d}] = x[C.kf{line_num:d}] * y[V.{pre_translocation}]"
            + (
                f" - x[C.kr{line_num:d}] * y[V.{post_translocation}]"
                if not is_unidirectional
                else ""
            )
        )
        self.kinetics.append(
            KineticInfo(
                (pre_translocation,),
                (post_translocation,),
                (),
                f"kf{line_num:d} * {pre_translocation}",
            )
        )
        if not is_unidirectional:
            self.kinetics.append(
                KineticInfo(
                    (post_translocation,),
                    (pre_translocation,),
                    (),
                    f"kr{line_num:d} * {post_translocation}",
                )
            )
        if float(pre_volume.strip(" ")) != float(post_volume.strip(" ")):
            self.reactions[
                -1
            ] = f"v[{line_num:d}] = " f"x[C.kf{line_num:d}] * y[V.{pre_translocation}]" + (
                f" - x[C.kr{line_num:d}] * "
                f"({post_volume.strip()} / {pre_volume.strip()}) * "
                f"y[V.{post_translocation}]"
                if not is_unidirectional
                else ""
            )
        counter_pre_translocation, counter_post_translocation = (0, 0)
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{pre_translocation}]" in eq:
                counter_pre_translocation += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif f"dydt[V.{post_translocation}]" in eq:
                counter_post_translocation += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
                if float(pre_volume.strip(" ")) != float(post_volume.strip(" ")):
                    self.differential_equations[
                        i
                    ] += f" * ({pre_volume.strip()} / {post_volume.strip()})"
        if counter_pre_translocation == 0:
            self.differential_equations.append(f"dydt[V.{pre_translocation}] = - v[{line_num:d}]")
        if counter_post_translocation == 0:
            self.differential_equations.append(f"dydt[V.{post_translocation}] = + v[{line_num:d}]")
            if float(pre_volume.strip(" ")) != float(post_volume.strip(" ")):
                self.differential_equations[
                    -1
                ] += f" * ({pre_volume.strip()} / {post_volume.strip()})"

    def state_transition(self, line_num: int, line: str) -> None:
        """
        This rule is applied only when any rule words are not detected.

        Examples
        --------
        'Reactant --> Product'
        'Reactant <--> Product'
        'E + S --> E + P'

        Notes
        -----
        * Parameters
            .. math:: kf (, kr)

        * Rate equation
            .. math:: v = kf * [Reactant] (- kr * [Product])

        * Differential equation
            .. math:: d[Reactant]/dt = - v
            .. math:: d[Product]/dt = + v

        """
        for arrow in self._available_arrows():
            if arrow in line:
                params_used = ["kf"] if arrow in self.fwd_arrows else ["kf", "kr"]
                break
        else:
            raise ArrowError(self._get_arrow_error_message(line_num) + ".")
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, *params_used
        )
        is_unidirectional: bool
        for arrow in self._available_arrows():
            if arrow in description[0]:
                is_unidirectional = True if arrow in self.fwd_arrows else False
                two_species = description[0].split(arrow)
                reactants = (
                    None
                    if len(two_species[0].split(" ")) > 1 and "+" not in two_species[0]
                    else two_species[0].replace(" ", "").split("+")
                )
                products = (
                    None
                    if len(two_species[1].split(" ")) > 1 and "+" not in two_species[1]
                    else two_species[1].replace(" ", "").split("+")
                )
                break
        else:
            raise ArrowError(self._get_arrow_error_message(line_num) + ".")
        if reactants is None or products is None:
            raise DetectionError(f"Unregistered words in line{line_num:d}: {line}")
        if list(set(reactants) & set(products)):
            modifiers = list(set(reactants) & set(products))
            for modifier in modifiers:
                reactants.remove(modifier)
                products.remove(modifier)
            self._set_species(*modifiers)
            modifiers_string = " * " + " * ".join([f"y[V.{modifier}]" for modifier in modifiers])
            modifiers_string_raw = " * " + " * ".join([f"{modifier}" for modifier in modifiers])
        else:
            modifiers = None
            modifiers_string = ""
        self._set_species(*reactants, *products)
        lefthand = "y[V." + "] * y[V.".join(reactants) + "]"
        lefthand_raw = " * ".join(reactants)
        righthand = "y[V." + "] * y[V.".join(products) + "]"
        righthand_raw = " * ".join(products)
        self.reactions.append(
            f"v[{line_num:d}] = "
            f"x[C.kf{line_num:d}] * "
            + lefthand
            + modifiers_string
            + (
                f" - x[C.kr{line_num:d}] * " + righthand + modifiers_string
                if not is_unidirectional
                else ""
            )
        )
        if modifiers is None:
            self.kinetics.append(
                KineticInfo(
                    tuple(reactants),
                    tuple(products),
                    (),
                    f"kf{line_num} * " + lefthand_raw,
                )
            )
        else:
            self.kinetics.append(
                KineticInfo(
                    tuple(reactants),
                    tuple(products),
                    tuple(modifiers),
                    f"kf{line_num} * " + lefthand_raw + modifiers_string_raw,
                )
            )
        if not is_unidirectional and modifiers is None:
            self.kinetics.append(
                KineticInfo(
                    tuple(products),
                    tuple(reactants),
                    (),
                    f"kr{line_num} * " + righthand_raw,
                )
            )
        elif not is_unidirectional and modifiers:
            self.kinetics.append(
                KineticInfo(
                    tuple(products),
                    tuple(reactants),
                    tuple(modifiers),
                    f"kr{line_num} * " + righthand_raw + modifiers_string_raw,
                )
            )

        counter_reactant = defaultdict(lambda: 0)
        counter_product = defaultdict(lambda: 0)
        for i, eq in enumerate(self.differential_equations):
            for reactant in reactants:
                if f"dydt[V.{reactant}]" in eq:
                    counter_reactant[reactant] += 1
                    self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            for product in products:
                if f"dydt[V.{product}]" in eq:
                    counter_product[product] += 1
                    self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        for reactant in reactants:
            if counter_reactant[reactant] == 0:
                self.differential_equations.append(f"dydt[V.{reactant}] = - v[{line_num:d}]")
        for product in products:
            if counter_product[product] == 0:
                self.differential_equations.append(f"dydt[V.{product}] = + v[{line_num:d}]")

    def user_defined(self, line_num: int, line: str) -> None:
        """
        Examples
        --------
        >>> '@rxn Reactant --> Product: define rate equation here'

        Notes
        -----
        * Use p[xxx] and u[xxx] for describing parameters and species, respectively.

        * Use '0' or '∅' for degradation/creation to/from nothing.

        * Differential equation
            .. math::

                d[Reactant]/dt = - v

                d[Product]/dt = + v
        """
        all_params = re.findall(r"p\[(.*?)\]", line)
        description = self._preprocessing(
            sys._getframe().f_code.co_name, line_num, line, *all_params
        )
        balance = description[0].strip()
        rate_equation = description[1].strip()
        for arrow in self.fwd_arrows:
            if arrow in balance:
                reactant, product = balance.split(arrow)
                if reactant in self.nothing:
                    self._set_species(product.strip())
                elif product in self.nothing:
                    self._set_species(product.strip())
                else:
                    self._set_species(reactant.strip(), product.strip())
                break
        else:
            raise ArrowError(f"line{line_num:d}: Use one of {', '.join(self.fwd_arrows)}.")
        rate_equation = (
            rate_equation.replace("p[", "x[C.").replace("u[", "y[V.").replace("^", "**")
        )
        self.reactions.append(f"v[{line_num:d}] = " + rate_equation.strip())
        modulators = (
            *list(
                set(
                    [
                        ent
                        for ent in re.findall(r"(?<=\[V.)(.+?)(?=\])", rate_equation)
                        if ent not in [reactant, product]
                    ]
                )
            ),
        )
        self.kinetics.append(
            KineticInfo(
                () if reactant in self.nothing else (reactant,),
                () if product in self.nothing else (product,),
                () if modulators is None else (modulators),
                rate_equation.replace("x[C.", "").replace("y[V.", "").replace("]", ""),
            )
        )
        counter_reactant = 0
        counter_product = 0
        for i, eq in enumerate(self.differential_equations):
            if f"dydt[V.{reactant}]" in eq and reactant not in self.nothing:
                counter_reactant += 1
                self.differential_equations[i] = eq + f" - v[{line_num:d}]"
            elif f"dydt[V.{product}]" in eq and product not in self.nothing:
                counter_product += 1
                self.differential_equations[i] = eq + f" + v[{line_num:d}]"
        if counter_reactant == 0 and reactant not in self.nothing:
            self.differential_equations.append(f"dydt[V.{reactant}] = - v[{line_num:d}]")
        if counter_product == 0 and product not in self.nothing:
            self.differential_equations.append(f"dydt[V.{product}] = + v[{line_num:d}]")

    def _extract_event(self, line_num: int, line: str):
        # About biochemical event
        if line.startswith("@rxn "):
            line = self._remove_prefix(line, "@rxn ")
            if line.count(":") != 1:
                raise SyntaxError(f"line{line_num:d}: Missing colon")
            else:
                self.user_defined(line_num, line)
        # About observables
        elif line.startswith("@obs "):
            line = self._remove_prefix(line, "@obs ")
            if line.count(":") != 1:
                raise SyntaxError(
                    f"line{line_num:d}: Missing colon\n"
                    "Should be `@obs <observable name>: <expression>`."
                )
            else:
                self.obs_desc.append(line.split(":"))
        # About simulation info.
        elif line.startswith("@sim "):
            line = self._remove_prefix(line, "@sim ")
            if line.count(":") != 1:
                raise SyntaxError(f"line{line_num:d}: Missing colon")
            else:
                if line.startswith("tspan"):
                    t_info = line.split(":")[-1].strip()
                    if "[" in t_info and "]" in t_info:
                        [t0, tf] = t_info.split("[")[-1].split("]")[0].split(",")
                        if t0.strip(" ").isdecimal() and tf.strip(" ").isdecimal():
                            self.sim_tspan.append(t0)
                            self.sim_tspan.append(tf)
                        else:
                            raise TypeError("@sim tspan: [t0, tf] must be a list of integers.")
                    else:
                        raise ValueError(
                            "`tspan` must be a two element vector [t0, tf] "
                            "specifying the initial and final times."
                        )
                elif line.startswith("unperturbed"):
                    self.sim_unperturbed += line.split(":")[-1].strip()
                elif line.startswith("condition "):
                    self.sim_conditions.append(self._remove_prefix(line, "condition ").split(":"))
                else:
                    raise ValueError(
                        f"(line{line_num:d}) Available options are: "
                        "'@sim tspan:', '@sim unperturbed:', or '@sim condition XXX:'."
                    )
        # Additional species
        elif line.startswith("@add "):
            line = self._remove_prefix(line, "@add ")
            if line.startswith("species "):
                line = self._remove_prefix(line, "species ")
                new_species = line.strip()
                if new_species not in self.species:
                    self._set_species(new_species)
                else:
                    raise NameError(f"{new_species} is already defined.")
            elif line.startswith("param "):
                line = self._remove_prefix(line, "param ")
                new_param = line.strip()
                if new_param not in self.parameters:
                    self._set_params(None, None, new_param)
                    self.param_excluded.append(new_param)
                else:
                    raise NameError(f"{new_param} is already defined.")
            else:
                raise ValueError(f"(line{line_num:d}) Must be either @add param or @add species.")
        else:
            raise ValueError("Available symbols are: @rxn, @add, @obs, @sim.")

    def create_ode(self) -> None:
        """
        Find a keyword in each line to identify the reaction rule and
        construct an ODE model.

        """
        with open(self.input_txt, encoding="utf-8") as f:
            lines = f.readlines()
        for line_num, line in enumerate(lines, start=1):
            # Remove double spaces
            while True:
                if "  " not in line:
                    break
                else:
                    line = line.replace("  ", " ")
            # Comment out
            line = line.split("#")[0].rstrip(" ")
            if not line.strip():
                # Skip blank lines
                continue
            elif lines.count(line) > 1:
                # Find duplicate lines
                raise DuplicateError(
                    f"Reaction '{line}' is duplicated in lines "
                    + ", ".join([str(i + 1) for i, rxn in enumerate(lines) if rxn == line])
                )
            elif line.startswith("@"):
                self._extract_event(line_num, line)
            # Detect reaction rule
            else:
                for reaction_rule, words in self.rule_words.items():
                    if any([self._remove_prepositions(word) in line for word in words]):
                        exec("self." + reaction_rule + "(line_num, line)")
                        break
                else:
                    self._raise_exception(line_num, line)



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    ## CYCE SYNTHESISDEGRADATION AND P27 BINDING/DISSOCIATION:
    "kscyce",
    "kdcyce",
    "kdcycee",
    "kdcycea",
    "kasse",
    "kdise",
    ## CYCA SYNTHESISDEGRADATION AND P27 BINDING/DISSOCIATION:
    "kscyca",
    "kdcyca",
    "kdcycac1",
    "kassa",
    "kdisa",
    ## P27 SYNTHESIS AND DEGRADATION:
    "ks27",
    "kd27",
    "kd27e",
    "kd27a",
    ## EMI1 SYNTHESIS AND DEGRADATION:
    "ksemi1",
    "kdemi1",
    ## CDH1 REGULATION:
    "Cdh1T",
    "kacdh1",
    "kicdh1e",
    "kicdh1a",
    "kasec",
    "kdiec",
    ## SKP2 SYNTHESIS AND DEGRADATION:
    "ksskp2",
    "kdskp2",
    "kdskp2c1",
    ## CDK INHIBITOR
    "Inhibitor",
]

NUM: int = len(NAMES)

Parameters = make_dataclass(
    cls_name="Parameters",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

C = Parameters(**name2idx)

del name2idx
